# -*- coding: utf-8 -*-
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
from ....interfaces.utility import IdentityInterface, Merge, Function
from ....pipeline import engine as pe
from ....interfaces.freesurfer import *
from .ba_maps import create_ba_maps_wf
from ....interfaces.io import DataGrabber


def create_AutoRecon3(name="AutoRecon3",
                      qcache=False,
                      plugin_args=None,
                      th3=True,
                      exvivo=True,
                      entorhinal=True,
                      fsvernum=5.3):

    # AutoRecon3
    # Workflow
    ar3_wf = pe.Workflow(name=name)

    # Input Node
    inputspec = pe.Node(
        IdentityInterface(fields=[
            'lh_inflated', 'rh_inflated', 'lh_smoothwm', 'rh_smoothwm',
            'lh_white', 'rh_white', 'lh_white_H', 'rh_white_H', 'lh_white_K',
            'rh_white_K', 'lh_cortex_label', 'rh_cortex_label', 'lh_orig',
            'rh_orig', 'lh_sulc', 'rh_sulc', 'lh_area', 'rh_area', 'lh_curv',
            'rh_curv', 'lh_orig_nofix', 'rh_orig_nofix', 'aseg_presurf',
            'brain_finalsurfs', 'wm', 'filled', 'brainmask', 'transform',
            'orig_mgz', 'rawavg', 'norm', 'lh_atlas', 'rh_atlas',
            'lh_classifier1', 'rh_classifier1', 'lh_classifier2',
            'rh_classifier2', 'lh_classifier3', 'rh_classifier3',
            'lookup_table', 'wm_lookup_table', 'src_subject_id',
            'src_subject_dir', 'color_table', 'num_threads'
        ]),
        name='inputspec')

    ar3_lh_wf1 = pe.Workflow(name="AutoRecon3_Left_1")
    ar3_rh_wf1 = pe.Workflow(name="AutoRecon3_Right_1")
    for hemisphere, hemi_wf in [('lh', ar3_lh_wf1), ('rh', ar3_rh_wf1)]:
        hemi_inputspec1 = pe.Node(
            IdentityInterface(fields=[
                'inflated', 'smoothwm', 'white', 'cortex_label', 'orig',
                'aseg_presurf', 'brain_finalsurfs', 'wm', 'filled', 'sphere',
                'sulc', 'area', 'curv', 'classifier', 'atlas', 'num_threads'
            ]),
            name="inputspec")

        # Spherical Inflation

        # Inflates the orig surface into a sphere while minimizing metric distortion.
        # This step is necessary in order to register the surface to the spherical
        # atlas (also known as the spherical morph). Calls mris_sphere. Creates
        # surf/?h.sphere. The -autorecon3 stage begins here.

        ar3_sphere = pe.Node(Sphere(), name="Spherical_Inflation")
        ar3_sphere.inputs.seed = 1234
        ar3_sphere.inputs.out_file = '{0}.sphere'.format(hemisphere)
        if plugin_args:
            ar3_sphere.plugin_args = plugin_args
        hemi_wf.connect([(hemi_inputspec1, ar3_sphere,
                          [('inflated', 'in_file'),
                           ('smoothwm', 'in_smoothwm'), ('num_threads',
                                                         'num_threads')])])

        # Ipsilateral Surface Registation (Spherical Morph)

        # Registers the orig surface to the spherical atlas through surf/?h.sphere.
        # The surfaces are first coarsely registered by aligning the large scale
        # folding patterns found in ?h.sulc and then fine tuned using the small-scale
        # patterns as in ?h.curv. Calls mris_register. Creates surf/?h.sphere.reg.

        ar3_surfreg = pe.Node(Register(), name="Surface_Registration")
        ar3_surfreg.inputs.out_file = '{0}.sphere.reg'.format(hemisphere)
        ar3_surfreg.inputs.curv = True
        hemi_wf.connect([(ar3_sphere, ar3_surfreg, [('out_file', 'in_surf')]),
                         (hemi_inputspec1, ar3_surfreg,
                          [('smoothwm', 'in_smoothwm'), ('sulc', 'in_sulc'),
                           ('atlas', 'target')])])

        # Jacobian

        # Computes how much the white surface was distorted in order to register to
        # the spherical atlas during the -surfreg step.

        ar3_jacobian = pe.Node(Jacobian(), name="Jacobian")
        ar3_jacobian.inputs.out_file = '{0}.jacobian_white'.format(hemisphere)
        hemi_wf.connect(
            [(hemi_inputspec1, ar3_jacobian, [('white', 'in_origsurf')]),
             (ar3_surfreg, ar3_jacobian, [('out_file', 'in_mappedsurf')])])

        # Average Curvature

        # Resamples the average curvature from the atlas to that of the subject.
        # Allows the user to display activity on the surface of an individual
        # with the folding pattern (ie, anatomy) of a group.

        ar3_paint = pe.Node(Paint(), name="Average_Curvature")
        ar3_paint.inputs.averages = 5
        ar3_paint.inputs.template_param = 6
        ar3_paint.inputs.out_file = "{0}.avg_curv".format(hemisphere)
        hemi_wf.connect([(ar3_surfreg, ar3_paint, [('out_file', 'in_surf')]),
                         (hemi_inputspec1, ar3_paint, [('atlas',
                                                        'template')])])

        # Cortical Parcellation

        # Assigns a neuroanatomical label to each location on the cortical
        # surface. Incorporates both geometric information derived from the
        # cortical model (sulcus and curvature), and neuroanatomical convention.

        ar3_parcellation = pe.Node(MRIsCALabel(), "Cortical_Parcellation")
        ar3_parcellation.inputs.seed = 1234
        ar3_parcellation.inputs.hemisphere = hemisphere
        ar3_parcellation.inputs.copy_inputs = True
        ar3_parcellation.inputs.out_file = "{0}.aparc.annot".format(hemisphere)
        if plugin_args:
            ar3_parcellation.plugin_args = plugin_args
        hemi_wf.connect(
            [(hemi_inputspec1, ar3_parcellation,
              [('smoothwm', 'smoothwm'), ('cortex_label', 'label'),
               ('aseg_presurf', 'aseg'), ('classifier', 'classifier'),
               ('curv', 'curv'), ('sulc', 'sulc'), ('num_threads',
                                                    'num_threads')]),
             (ar3_surfreg, ar3_parcellation, [('out_file', 'canonsurf')])])

        # Pial Surface

        ar3_pial = pe.Node(MakeSurfaces(), name="Make_Pial_Surface")
        ar3_pial.inputs.mgz = True
        ar3_pial.inputs.hemisphere = hemisphere
        ar3_pial.inputs.copy_inputs = True

        if fsvernum < 6:
            ar3_pial.inputs.white = 'NOWRITE'
            hemi_wf.connect(hemi_inputspec1, 'white', ar3_pial, 'in_white')
        else:
            ar3_pial.inputs.no_white = True
            hemi_wf.connect([(hemi_inputspec1, ar3_pial,
                              [('white', 'orig_pial'), ('white',
                                                        'orig_white')])])

        hemi_wf.connect(
            [(hemi_inputspec1, ar3_pial,
              [('wm', 'in_wm'), ('orig', 'in_orig'), ('filled', 'in_filled'),
               ('brain_finalsurfs', 'in_T1'), ('aseg_presurf', 'in_aseg')]),
             (ar3_parcellation, ar3_pial, [('out_file', 'in_label')])])

        # Surface Volume
        """
        Creates the ?h.volume file by first creating the ?h.mid.area file by
        adding ?h.area(.white) to ?h.area.pial, then dividing by two. Then ?h.volume
        is created by multiplying ?.mid.area with ?h.thickness.
        """

        ar3_add = pe.Node(MRIsCalc(), name="Add_Pial_Area")
        ar3_add.inputs.action = "add"
        ar3_add.inputs.out_file = '{0}.area.mid'.format(hemisphere)
        hemi_wf.connect([
            (ar3_pial, ar3_add, [('out_area', 'in_file2')]),
            (hemi_inputspec1, ar3_add, [('area', 'in_file1')]),
        ])

        ar3_divide = pe.Node(MRIsCalc(), name="Mid_Pial")
        ar3_divide.inputs.action = "div"
        ar3_divide.inputs.in_int = 2
        ar3_divide.inputs.out_file = '{0}.area.mid'.format(hemisphere)
        hemi_wf.connect([
            (ar3_add, ar3_divide, [('out_file', 'in_file1')]),
        ])

        ar3_volume = pe.Node(MRIsCalc(), name="Calculate_Volume")
        ar3_volume.inputs.action = "mul"
        ar3_volume.inputs.out_file = '{0}.volume'.format(hemisphere)
        hemi_wf.connect([
            (ar3_divide, ar3_volume, [('out_file', 'in_file1')]),
            (ar3_pial, ar3_volume, [('out_thickness', 'in_file2')]),
        ])

        # Connect the inputs
        ar3_wf.connect(
            [(inputspec, hemi_wf,
              [('{0}_inflated'.format(hemisphere), 'inputspec.inflated'),
               ('{0}_smoothwm'.format(hemisphere),
                'inputspec.smoothwm'), ('{0}_white'.format(hemisphere),
                                        'inputspec.white'),
               ('{0}_cortex_label'.format(hemisphere),
                'inputspec.cortex_label'), ('{0}_orig'.format(hemisphere),
                                            'inputspec.orig'),
               ('{0}_sulc'.format(hemisphere),
                'inputspec.sulc'), ('{0}_area'.format(hemisphere),
                                    'inputspec.area'),
               ('{0}_curv'.format(hemisphere),
                'inputspec.curv'), ('aseg_presurf', 'inputspec.aseg_presurf'),
               ('brain_finalsurfs',
                'inputspec.brain_finalsurfs'), ('wm', 'inputspec.wm'),
               ('filled', 'inputspec.filled'), ('{0}_atlas'.format(hemisphere),
                                                'inputspec.atlas'),
               ('{0}_classifier1'.format(hemisphere),
                'inputspec.classifier'), ('num_threads',
                                          'inputspec.num_threads')])])

        # Workflow1 Outputs
        hemi_outputs1 = [
            'sphere', 'sphere_reg', 'jacobian_white', 'avg_curv',
            'aparc_annot', 'area_pial', 'curv_pial', 'pial', 'thickness_pial',
            'area_mid', 'volume'
        ]
        hemi_outputspec1 = pe.Node(
            IdentityInterface(fields=hemi_outputs1), name="outputspec")
        hemi_wf.connect([(ar3_pial, hemi_outputspec1, [
            ('out_pial', 'pial'), ('out_curv', 'curv_pial'),
            ('out_area', 'area_pial'), ('out_thickness', 'thickness_pial')
        ]), (ar3_divide, hemi_outputspec1,
             [('out_file', 'area_mid')]), (ar3_volume, hemi_outputspec1,
                                           [('out_file', 'volume')]),
                         (ar3_parcellation, hemi_outputspec1,
                          [('out_file', 'aparc_annot')]),
                         (ar3_jacobian, hemi_outputspec1,
                          [('out_file',
                            'jacobian_white')]), (ar3_paint, hemi_outputspec1,
                                                  [('out_file', 'avg_curv')]),
                         (ar3_surfreg, hemi_outputspec1,
                          [('out_file',
                            'sphere_reg')]), (ar3_sphere, hemi_outputspec1,
                                              [('out_file', 'sphere')])])

    # Cortical Ribbon Mask
    """
    Creates binary volume masks of the cortical ribbon
    ie, each voxel is either a 1 or 0 depending upon whether it falls in the ribbon or not.
    """
    volume_mask = pe.Node(VolumeMask(), name="Mask_Ribbon")
    volume_mask.inputs.left_whitelabel = 2
    volume_mask.inputs.left_ribbonlabel = 3
    volume_mask.inputs.right_whitelabel = 41
    volume_mask.inputs.right_ribbonlabel = 42
    volume_mask.inputs.save_ribbon = True
    volume_mask.inputs.copy_inputs = True

    ar3_wf.connect([
        (inputspec, volume_mask, [('lh_white', 'lh_white'), ('rh_white',
                                                             'rh_white')]),
        (ar3_lh_wf1, volume_mask, [('outputspec.pial', 'lh_pial')]),
        (ar3_rh_wf1, volume_mask, [('outputspec.pial', 'rh_pial')]),
    ])

    if fsvernum >= 6:
        ar3_wf.connect([(inputspec, volume_mask, [('aseg_presurf',
                                                   'in_aseg')])])
    else:
        ar3_wf.connect([(inputspec, volume_mask, [('aseg_presurf', 'aseg')])])

    ar3_lh_wf2 = pe.Workflow(name="AutoRecon3_Left_2")
    ar3_rh_wf2 = pe.Workflow(name="AutoRecon3_Right_2")

    for hemisphere, hemiwf2 in [('lh', ar3_lh_wf2), ('rh', ar3_rh_wf2)]:
        if hemisphere == 'lh':
            hemiwf1 = ar3_lh_wf1
        else:
            hemiwf1 = ar3_rh_wf1

        hemi_inputs2 = [
            'wm',
            'lh_white',
            'rh_white',
            'transform',
            'brainmask',
            'aseg_presurf',
            'cortex_label',
            'lh_pial',
            'rh_pial',
            'thickness',
            'aparc_annot',
            'ribbon',
            'smoothwm',
            'sphere_reg',
            'orig_mgz',
            'rawavg',
            'curv',
            'sulc',
            'classifier2',
            'classifier3',
        ]

        hemi_inputspec2 = pe.Node(
            IdentityInterface(fields=hemi_inputs2), name="inputspec")

        # Parcellation Statistics
        """
        Runs mris_anatomical_stats to create a summary table of cortical parcellation statistics for each structure, including
        structure name
        number of vertices
        total surface area (mm^2)
        total gray matter volume (mm^3)
        average cortical thickness (mm)
        standard error of cortical thicknessr (mm)
        integrated rectified mean curvature
        integrated rectified Gaussian curvature
        folding index
        intrinsic curvature index.
        """
        parcellation_stats_white = pe.Node(
            ParcellationStats(),
            name="Parcellation_Stats_{0}_White".format(hemisphere))
        parcellation_stats_white.inputs.mgz = True
        parcellation_stats_white.inputs.th3 = th3
        parcellation_stats_white.inputs.tabular_output = True
        parcellation_stats_white.inputs.surface = 'white'
        parcellation_stats_white.inputs.hemisphere = hemisphere
        parcellation_stats_white.inputs.out_color = 'aparc.annot.ctab'
        parcellation_stats_white.inputs.out_table = '{0}.aparc.stats'.format(
            hemisphere)
        parcellation_stats_white.inputs.copy_inputs = True

        hemiwf2.connect([
            (hemi_inputspec2, parcellation_stats_white, [
                ('wm', 'wm'),
                ('lh_white', 'lh_white'),
                ('rh_white', 'rh_white'),
                ('transform', 'transform'),
                ('brainmask', 'brainmask'),
                ('aseg_presurf', 'aseg'),
                ('cortex_label', 'in_cortex'),
                ('cortex_label', 'cortex_label'),
                ('lh_pial', 'lh_pial'),
                ('rh_pial', 'rh_pial'),
                ('thickness', 'thickness'),
                ('aparc_annot', 'in_annotation'),
                ('ribbon', 'ribbon'),
            ]),
        ])

        parcellation_stats_pial = pe.Node(
            ParcellationStats(),
            name="Parcellation_Stats_{0}_Pial".format(hemisphere))
        parcellation_stats_pial.inputs.mgz = True
        parcellation_stats_pial.inputs.th3 = th3
        parcellation_stats_pial.inputs.tabular_output = True
        parcellation_stats_pial.inputs.surface = 'pial'
        parcellation_stats_pial.inputs.hemisphere = hemisphere
        parcellation_stats_pial.inputs.copy_inputs = True
        parcellation_stats_pial.inputs.out_color = 'aparc.annot.ctab'
        parcellation_stats_pial.inputs.out_table = '{0}.aparc.pial.stats'.format(
            hemisphere)

        hemiwf2.connect([
            (hemi_inputspec2, parcellation_stats_pial, [
                ('wm', 'wm'),
                ('lh_white', 'lh_white'),
                ('rh_white', 'rh_white'),
                ('transform', 'transform'),
                ('brainmask', 'brainmask'),
                ('aseg_presurf', 'aseg'),
                ('cortex_label', 'cortex_label'),
                ('cortex_label', 'in_cortex'),
                ('lh_pial', 'lh_pial'),
                ('rh_pial', 'rh_pial'),
                ('thickness', 'thickness'),
                ('aparc_annot', 'in_annotation'),
                ('ribbon', 'ribbon'),
            ]),
        ])

        # Cortical Parcellation 2
        cortical_parcellation_2 = pe.Node(
            MRIsCALabel(),
            name="Cortical_Parcellation_{0}_2".format(hemisphere))
        cortical_parcellation_2.inputs.out_file = '{0}.aparc.a2009s.annot'.format(
            hemisphere)
        cortical_parcellation_2.inputs.seed = 1234
        cortical_parcellation_2.inputs.copy_inputs = True
        cortical_parcellation_2.inputs.hemisphere = hemisphere

        hemiwf2.connect([(hemi_inputspec2, cortical_parcellation_2,
                          [('smoothwm', 'smoothwm'), ('aseg_presurf', 'aseg'),
                           ('cortex_label', 'label'), ('sphere_reg',
                                                       'canonsurf'), ('curv',
                                                                      'curv'),
                           ('sulc', 'sulc'), ('classifier2', 'classifier')])])

        # Parcellation Statistics 2
        parcellation_stats_white_2 = parcellation_stats_white.clone(
            name="Parcellation_Statistics_{0}_2".format(hemisphere))
        parcellation_stats_white_2.inputs.hemisphere = hemisphere
        parcellation_stats_white_2.inputs.out_color = 'aparc.annot.a2009s.ctab'
        parcellation_stats_white_2.inputs.out_table = '{0}.aparc.a2009s.stats'.format(
            hemisphere)
        hemiwf2.connect([(hemi_inputspec2, parcellation_stats_white_2, [
            ('wm', 'wm'),
            ('lh_white', 'lh_white'),
            ('rh_white', 'rh_white'),
            ('transform', 'transform'),
            ('brainmask', 'brainmask'),
            ('aseg_presurf', 'aseg'),
            ('cortex_label', 'cortex_label'),
            ('cortex_label', 'in_cortex'),
            ('lh_pial', 'lh_pial'),
            ('rh_pial', 'rh_pial'),
            ('thickness', 'thickness'),
            ('ribbon', 'ribbon'),
        ]), (cortical_parcellation_2, parcellation_stats_white_2,
             [('out_file', 'in_annotation')])])

        # Cortical Parcellation 3
        cortical_parcellation_3 = pe.Node(
            MRIsCALabel(),
            name="Cortical_Parcellation_{0}_3".format(hemisphere))
        cortical_parcellation_3.inputs.out_file = '{0}.aparc.DKTatlas40.annot'.format(
            hemisphere)
        cortical_parcellation_3.inputs.hemisphere = hemisphere
        cortical_parcellation_3.inputs.seed = 1234
        cortical_parcellation_3.inputs.copy_inputs = True
        hemiwf2.connect([(hemi_inputspec2, cortical_parcellation_3,
                          [('smoothwm', 'smoothwm'), ('aseg_presurf', 'aseg'),
                           ('cortex_label', 'label'), ('sphere_reg',
                                                       'canonsurf'), ('curv',
                                                                      'curv'),
                           ('sulc', 'sulc'), ('classifier3', 'classifier')])])

        # Parcellation Statistics 3
        parcellation_stats_white_3 = parcellation_stats_white.clone(
            name="Parcellation_Statistics_{0}_3".format(hemisphere))
        parcellation_stats_white_3.inputs.out_color = 'aparc.annot.DKTatlas40.ctab'
        parcellation_stats_white_3.inputs.out_table = '{0}.aparc.DKTatlas40.stats'.format(
            hemisphere)
        parcellation_stats_white_3.inputs.hemisphere = hemisphere

        hemiwf2.connect([(hemi_inputspec2, parcellation_stats_white_3, [
            ('wm', 'wm'),
            ('lh_white', 'lh_white'),
            ('rh_white', 'rh_white'),
            ('transform', 'transform'),
            ('brainmask', 'brainmask'),
            ('aseg_presurf', 'aseg'),
            ('cortex_label', 'cortex_label'),
            ('cortex_label', 'in_cortex'),
            ('lh_pial', 'lh_pial'),
            ('rh_pial', 'rh_pial'),
            ('thickness', 'thickness'),
            ('ribbon', 'ribbon'),
        ]), (cortical_parcellation_3, parcellation_stats_white_3,
             [('out_file', 'in_annotation')])])

        # WM/GM Contrast
        contrast = pe.Node(
            Contrast(), name="WM_GM_Contrast_{0}".format(hemisphere))
        contrast.inputs.hemisphere = hemisphere
        contrast.inputs.copy_inputs = True

        hemiwf2.connect([
            (hemi_inputspec2, contrast, [
                ('orig_mgz', 'orig'),
                ('rawavg', 'rawavg'),
                ('{0}_white'.format(hemisphere), 'white'),
                ('cortex_label', 'cortex'),
                ('aparc_annot', 'annotation'),
                ('thickness', 'thickness'),
            ]),
        ])

        hemi_outputs2 = [
            'aparc_annot_ctab',
            'aparc_stats',
            'aparc_pial_stats',
            'aparc_a2009s_annot',
            'aparc_a2009s_annot_ctab',
            'aparc_a2009s_annot_stats',
            'aparc_DKTatlas40_annot',
            'aparc_DKTatlas40_annot_ctab',
            'aparc_DKTatlas40_annot_stats',
            'wg_pct_mgh',
            'wg_pct_stats',
            'pctsurfcon_log',
        ]
        hemi_outputspec2 = pe.Node(
            IdentityInterface(fields=hemi_outputs2), name="outputspec")

        hemiwf2.connect([
            (contrast, hemi_outputspec2,
             [('out_contrast', 'wg_pct_mgh'), ('out_stats', 'wg_pct_stats'),
              ('out_log', 'pctsurfcon_log')]),
            (parcellation_stats_white_3, hemi_outputspec2,
             [('out_color', 'aparc_DKTatlas40_annot_ctab'),
              ('out_table', 'aparc_DKTatlas40_annot_stats')]),
            (cortical_parcellation_3, hemi_outputspec2,
             [('out_file', 'aparc_DKTatlas40_annot')]),
            (parcellation_stats_white_2, hemi_outputspec2,
             [('out_color', 'aparc_a2009s_annot_ctab'),
              ('out_table', 'aparc_a2009s_annot_stats')]),
            (cortical_parcellation_2, hemi_outputspec2,
             [('out_file', 'aparc_a2009s_annot')]),
            (parcellation_stats_white, hemi_outputspec2,
             [('out_color', 'aparc_annot_ctab'), ('out_table',
                                                  'aparc_stats')]),
            (parcellation_stats_pial, hemi_outputspec2,
             [('out_table', 'aparc_pial_stats')]),
        ])
        # connect inputs to hemisphere2 workflow
        ar3_wf.connect([
            (inputspec, hemiwf2, [
                ('wm', 'inputspec.wm'),
                ('lh_white', 'inputspec.lh_white'),
                ('rh_white', 'inputspec.rh_white'),
                ('transform', 'inputspec.transform'),
                ('brainmask', 'inputspec.brainmask'),
                ('aseg_presurf', 'inputspec.aseg_presurf'),
                ('{0}_cortex_label'.format(hemisphere),
                 'inputspec.cortex_label'),
                ('{0}_smoothwm'.format(hemisphere), 'inputspec.smoothwm'),
                ('orig_mgz', 'inputspec.orig_mgz'),
                ('rawavg', 'inputspec.rawavg'),
                ('{0}_curv'.format(hemisphere), 'inputspec.curv'),
                ('{0}_sulc'.format(hemisphere), 'inputspec.sulc'),
                ('{0}_classifier2'.format(hemisphere),
                 'inputspec.classifier2'),
                ('{0}_classifier3'.format(hemisphere),
                 'inputspec.classifier3'),
            ]),
            (ar3_lh_wf1, hemiwf2, [('outputspec.pial', 'inputspec.lh_pial')]),
            (ar3_rh_wf1, hemiwf2, [('outputspec.pial', 'inputspec.rh_pial')]),
            (hemiwf1, hemiwf2,
             [('outputspec.thickness_pial', 'inputspec.thickness'),
              ('outputspec.aparc_annot', 'inputspec.aparc_annot'),
              ('outputspec.sphere_reg', 'inputspec.sphere_reg')]),
            (volume_mask, hemiwf2, [('out_ribbon', 'inputspec.ribbon')]),
        ])
        # End hemisphere2 workflow

    # APARC to ASEG
    # Adds information from the ribbon into the aseg.mgz (volume parcellation).
    aparc_2_aseg = pe.Node(Aparc2Aseg(), name="Aparc2Aseg")
    aparc_2_aseg.inputs.volmask = True
    aparc_2_aseg.inputs.copy_inputs = True
    aparc_2_aseg.inputs.out_file = "aparc+aseg.mgz"
    ar3_wf.connect([(inputspec, aparc_2_aseg, [
        ('lh_white', 'lh_white'),
        ('rh_white', 'rh_white'),
    ]), (ar3_lh_wf1, aparc_2_aseg, [
        ('outputspec.pial', 'lh_pial'),
        ('outputspec.aparc_annot', 'lh_annotation'),
    ]), (ar3_rh_wf1, aparc_2_aseg, [
        ('outputspec.pial', 'rh_pial'),
        ('outputspec.aparc_annot', 'rh_annotation'),
    ]), (volume_mask, aparc_2_aseg, [
        ('rh_ribbon', 'rh_ribbon'),
        ('lh_ribbon', 'lh_ribbon'),
        ('out_ribbon', 'ribbon'),
    ])])
    if fsvernum < 6:
        ar3_wf.connect([(inputspec, aparc_2_aseg, [('aseg_presurf', 'aseg')])])
    else:
        # Relabel Hypointensities
        relabel_hypos = pe.Node(
            RelabelHypointensities(), name="Relabel_Hypointensities")
        relabel_hypos.inputs.out_file = 'aseg.presurf.hypos.mgz'
        ar3_wf.connect([(inputspec, relabel_hypos,
                         [('aseg_presurf', 'aseg'), ('lh_white', 'lh_white'),
                          ('rh_white', 'rh_white')])])
        ar3_wf.connect([(relabel_hypos, aparc_2_aseg, [('out_file', 'aseg')])])

    aparc_2_aseg_2009 = pe.Node(Aparc2Aseg(), name="Aparc2Aseg_2009")
    aparc_2_aseg_2009.inputs.volmask = True
    aparc_2_aseg_2009.inputs.a2009s = True
    aparc_2_aseg_2009.inputs.copy_inputs = True
    aparc_2_aseg_2009.inputs.out_file = "aparc.a2009s+aseg.mgz"
    ar3_wf.connect([(inputspec, aparc_2_aseg_2009, [
        ('lh_white', 'lh_white'),
        ('rh_white', 'rh_white'),
    ]), (ar3_lh_wf1, aparc_2_aseg_2009, [
        ('outputspec.pial', 'lh_pial'),
    ]), (ar3_lh_wf2, aparc_2_aseg_2009, [('outputspec.aparc_a2009s_annot',
                                          'lh_annotation')]),
                    (ar3_rh_wf2, aparc_2_aseg_2009,
                     [('outputspec.aparc_a2009s_annot',
                       'rh_annotation')]), (ar3_rh_wf1, aparc_2_aseg_2009, [
                           ('outputspec.pial', 'rh_pial'),
                       ]), (volume_mask, aparc_2_aseg_2009,
                            [('rh_ribbon', 'rh_ribbon'),
                             ('lh_ribbon', 'lh_ribbon'), ('out_ribbon',
                                                          'ribbon')])])

    if fsvernum >= 6:
        apas_2_aseg = pe.Node(Apas2Aseg(), name="Apas_2_Aseg")
        ar3_wf.connect([(aparc_2_aseg, apas_2_aseg, [('out_file', 'in_file')]),
                        (relabel_hypos, aparc_2_aseg_2009, [('out_file',
                                                             'aseg')])])
    else:
        # aseg.mgz gets edited in place, so we'll copy and pass it to the
        # outputspec once aparc_2_aseg has completed
        def out_aseg(in_aparcaseg, in_aseg, out_file):
            import shutil
            import os
            out_file = os.path.abspath(out_file)
            shutil.copy(in_aseg, out_file)
            return out_file

        apas_2_aseg = pe.Node(
            Function(['in_aparcaseg', 'in_aseg', 'out_file'], ['out_file'],
                     out_aseg),
            name="Aseg")
        ar3_wf.connect(
            [(aparc_2_aseg, apas_2_aseg, [('out_file', 'in_aparcaseg')]),
             (inputspec, apas_2_aseg, [('aseg_presurf', 'in_aseg')]),
             (inputspec, aparc_2_aseg_2009, [('aseg_presurf', 'aseg')])])

    apas_2_aseg.inputs.out_file = "aseg.mgz"

    # Segmentation Stats
    """
    Computes statistics on the segmented subcortical structures found in
    mri/aseg.mgz. Writes output to file stats/aseg.stats.
    """

    segstats = pe.Node(SegStatsReconAll(), name="Segmentation_Statistics")
    segstats.inputs.empty = True
    segstats.inputs.brain_vol = 'brain-vol-from-seg'
    segstats.inputs.exclude_ctx_gm_wm = True
    segstats.inputs.supratent = True
    segstats.inputs.subcort_gm = True
    segstats.inputs.etiv = True
    segstats.inputs.wm_vol_from_surf = True
    segstats.inputs.cortex_vol_from_surf = True
    segstats.inputs.total_gray = True
    segstats.inputs.euler = True
    segstats.inputs.exclude_id = 0
    segstats.inputs.intensity_units = "MR"
    segstats.inputs.summary_file = 'aseg.stats'
    segstats.inputs.copy_inputs = True

    ar3_wf.connect([
        (apas_2_aseg, segstats, [('out_file', 'segmentation_file')]),
        (inputspec, segstats, [
            ('lh_white', 'lh_white'),
            ('rh_white', 'rh_white'),
            ('transform', 'transform'),
            ('norm', 'in_intensity'),
            ('norm', 'partial_volume_file'),
            ('brainmask', 'brainmask_file'),
            ('lh_orig_nofix', 'lh_orig_nofix'),
            ('rh_orig_nofix', 'rh_orig_nofix'),
            ('lookup_table', 'color_table_file'),
        ]),
        (volume_mask, segstats, [('out_ribbon', 'ribbon')]),
        (ar3_lh_wf1, segstats, [
            ('outputspec.pial', 'lh_pial'),
        ]),
        (ar3_rh_wf1, segstats, [
            ('outputspec.pial', 'rh_pial'),
        ]),
    ])

    if fsvernum >= 6:
        ar3_wf.connect(inputspec, 'aseg_presurf', segstats, 'presurf_seg')
    else:
        ar3_wf.connect(inputspec, 'aseg_presurf', segstats, 'aseg')

    # White Matter Parcellation

    # Adds WM Parcellation info into the aseg and computes stat.

    wm_parcellation = pe.Node(Aparc2Aseg(), name="WM_Parcellation")
    wm_parcellation.inputs.volmask = True
    wm_parcellation.inputs.label_wm = True
    wm_parcellation.inputs.hypo_wm = True
    wm_parcellation.inputs.rip_unknown = True
    wm_parcellation.inputs.copy_inputs = True
    wm_parcellation.inputs.out_file = "wmparc.mgz"

    ar3_wf.connect([(inputspec, wm_parcellation, [
        ('lh_white', 'lh_white'),
        ('rh_white', 'rh_white'),
    ]), (ar3_lh_wf1, wm_parcellation, [
        ('outputspec.pial', 'lh_pial'),
        ('outputspec.aparc_annot', 'lh_annotation'),
    ]), (ar3_rh_wf1, wm_parcellation, [
        ('outputspec.pial', 'rh_pial'),
        ('outputspec.aparc_annot', 'rh_annotation'),
    ]), (volume_mask, wm_parcellation, [
        ('rh_ribbon', 'rh_ribbon'),
        ('lh_ribbon', 'lh_ribbon'),
        ('out_ribbon', 'ribbon'),
    ]), (apas_2_aseg, wm_parcellation, [('out_file', 'aseg')]),
                    (aparc_2_aseg, wm_parcellation, [('out_file', 'ctxseg')])])

    if fsvernum < 6:
        ar3_wf.connect([(inputspec, wm_parcellation, [('filled', 'filled')])])

    # White Matter Segmentation Stats

    wm_segstats = pe.Node(
        SegStatsReconAll(), name="WM_Segmentation_Statistics")
    wm_segstats.inputs.intensity_units = "MR"
    wm_segstats.inputs.wm_vol_from_surf = True
    wm_segstats.inputs.etiv = True
    wm_segstats.inputs.copy_inputs = True
    wm_segstats.inputs.exclude_id = 0
    wm_segstats.inputs.summary_file = "wmparc.stats"

    ar3_wf.connect([
        (wm_parcellation, wm_segstats, [('out_file', 'segmentation_file')]),
        (inputspec, wm_segstats, [
            ('lh_white', 'lh_white'),
            ('rh_white', 'rh_white'),
            ('transform', 'transform'),
            ('norm', 'in_intensity'),
            ('norm', 'partial_volume_file'),
            ('brainmask', 'brainmask_file'),
            ('lh_orig_nofix', 'lh_orig_nofix'),
            ('rh_orig_nofix', 'rh_orig_nofix'),
            ('wm_lookup_table', 'color_table_file'),
        ]),
        (volume_mask, wm_segstats, [('out_ribbon', 'ribbon')]),
        (ar3_lh_wf1, wm_segstats, [
            ('outputspec.pial', 'lh_pial'),
        ]),
        (ar3_rh_wf1, wm_segstats, [
            ('outputspec.pial', 'rh_pial'),
        ]),
    ])

    if fsvernum >= 6:
        ar3_wf.connect(inputspec, 'aseg_presurf', wm_segstats, 'presurf_seg')
    else:
        ar3_wf.connect(inputspec, 'aseg_presurf', wm_segstats, 'aseg')

    # add brodman area maps to the workflow
    ba_WF, ba_outputs = create_ba_maps_wf(
        th3=th3, exvivo=exvivo, entorhinal=entorhinal)

    ar3_wf.connect([(ar3_lh_wf1, ba_WF, [
        ('outputspec.sphere_reg', 'inputspec.lh_sphere_reg'),
        ('outputspec.thickness_pial', 'inputspec.lh_thickness'),
        ('outputspec.pial', 'inputspec.lh_pial'),
    ]), (ar3_rh_wf1, ba_WF, [
        ('outputspec.sphere_reg', 'inputspec.rh_sphere_reg'),
        ('outputspec.thickness_pial', 'inputspec.rh_thickness'),
        ('outputspec.pial', 'inputspec.rh_pial'),
    ]), (inputspec, ba_WF, [
        ('lh_white', 'inputspec.lh_white'),
        ('rh_white', 'inputspec.rh_white'),
        ('transform', 'inputspec.transform'),
        ('aseg_presurf', 'inputspec.aseg'),
        ('brainmask', 'inputspec.brainmask'),
        ('wm', 'inputspec.wm'),
        ('lh_orig', 'inputspec.lh_orig'),
        ('rh_orig', 'inputspec.rh_orig'),
        ('lh_cortex_label', 'inputspec.lh_cortex_label'),
        ('rh_cortex_label', 'inputspec.rh_cortex_label'),
        ('src_subject_dir', 'inputspec.src_subject_dir'),
        ('src_subject_id', 'inputspec.src_subject_id'),
        ('color_table', 'inputspec.color_table'),
    ]), (volume_mask, ba_WF, [('out_ribbon', 'inputspec.ribbon')])])

    if qcache:
        source_inputs = ['lh_sphere_reg', 'rh_sphere_reg']
        source_subject = pe.Node(
            DataGrabber(outfields=source_inputs),
            name="{0}_srcsubject".format(hemisphere))
        source_subject.inputs.template = '*'
        source_subject.inputs.sort_filelist = False
        source_subject.inputs.field_template = dict(
            lh_sphere_reg='surf/lh.sphere.reg',
            rh_sphere_reg='surf/rh.sphere.reg')

        qcache_wf = pe.Workflow("QCache")

        measurements = [
            'thickness', 'area', 'area.pial', 'volume', 'curv', 'sulc',
            'white.K', 'white.H', 'jacobian_white', 'w-g.pct.mgh'
        ]

        qcache_inputs = list()
        for source_file in source_inputs:
            qcache_inputs.append('source_' + source_file)
        qcache_config = dict()
        qcache_outputs = list()
        for hemisphere in ['lh', 'rh']:
            qcache_config[hemisphere] = dict()
            for meas_name in measurements:
                qcache_config[hemisphere][meas_name] = dict()

                if meas_name == 'thickness':
                    meas_file = hemisphere + '_' + meas_name + '_pial'
                else:
                    meas_file = hemisphere + '_' + meas_name.replace(
                        '.', '_').replace('-', '')
                qcache_inputs.append(meas_file)

                preproc_name = "Preproc_{0}".format(meas_file)
                preproc_out = '{0}.{1}.{2}.mgh'.format(
                    hemisphere, meas_name, config['src_subject_id'])
                preproc_out_name = preproc_out.replace('.', '_')
                qcache_config[hemisphere][meas_name]['preproc'] = dict(
                    infile=meas_file,
                    name=preproc_name,
                    out=preproc_out,
                    out_name=preproc_out_name)
                qcache_outputs.append(preproc_out_name)

                qcache_config[hemisphere][meas_name]['smooth'] = dict()
                for value in range(0, 26, 5):
                    smooth_name = "Smooth_{0}_{1}".format(meas_file, value)
                    smooth_out = "{0}.{1}.fwhm{2}.{3}.mgh".format(
                        hemisphere, meas_name, value, config['src_subject_id'])
                    smooth_out_name = smooth_out.replace('.', '_')
                    qcache_config[hemisphere][meas_name]['smooth'][
                        value] = dict(
                            name=smooth_name,
                            out=smooth_out,
                            out_name=smooth_out_name)
                    qcache_outputs.append(smooth_out_name)

            qcache_inputs.append(hemisphere + '_sphere_reg')

        qcache_inputspec = pe.Node(
            IdentityInterface(fields=qcache_inputs), name="inputspec")

        qcache_outputspec = pe.Node(
            IdentityInterface(fields=qcache_outputs), name="outputspec")

        for hemi in qcache_config.iterkeys():
            for meas_config in qcache_config[hemi].itervalues():
                preprocess = pe.Node(
                    MRISPreprocReconAll(), name=meas_config['preproc']['name'])
                target_id = config['src_subject_id']
                preprocess.inputs.out_file = meas_config['preproc']['out']
                preprocess.inputs.target = target_id
                preprocess.inputs.hemi = hemi
                preprocess.inputs.copy_inputs = True

                qcache_merge = pe.Node(
                    Merge(2),
                    name="Merge{0}".format(meas_config['preproc']['name']))

                qcache_wf.connect([
                    (qcache_inputspec, qcache_merge,
                     [('lh_sphere_reg', 'in1'), ('rh_sphere_reg', 'in2')]),
                    (qcache_inputspec, preprocess,
                     [(meas_config['preproc']['infile'], 'surf_measure_file'),
                      ('source_lh_sphere_reg', 'lh_surfreg_target'),
                      ('source_rh_sphere_reg', 'rh_surfreg_target')]),
                    (qcache_merge, preprocess, [('out', 'surfreg_files')]),
                    (preprocess, qcache_outputspec,
                     [('out_file', meas_config['preproc']['out_name'])]),
                ])

                for value, val_config in meas_config['smooth'].iteritems():
                    surf2surf = pe.Node(
                        SurfaceSmooth(), name=val_config['name'])
                    surf2surf.inputs.fwhm = value
                    surf2surf.inputs.cortex = True
                    surf2surf.inputs.subject_id = target_id
                    surf2surf.inputs.hemi = hemisphere
                    surf2surf.inputs.out_file = val_config['out']
                    qcache_wf.connect(
                        [(preprocess, surf2surf, [('out_file', 'in_file')]),
                         (surf2surf, qcache_outputspec,
                          [('out_file', val_config['out_name'])])])

        # connect qcache inputs
        ar3_wf.connect([
            (inputspec, qcache_wf,
             [('lh_curv', 'inputspec.lh_curv'), ('rh_curv',
                                                 'inputspec.rh_curv'),
              ('lh_sulc', 'inputspec.lh_sulc'), ('rh_sulc',
                                                 'inputspec.rh_sulc'),
              ('lh_white_K', 'inputspec.lh_white_K'), ('rh_white_K',
                                                       'inputspec.rh_white_K'),
              ('lh_area', 'inputspec.lh_area'), ('rh_area',
                                                 'inputspec.rh_area')]),
            (ar3_lh_wf1, qcache_wf,
             [('outputspec.thickness_pial', 'inputspec.lh_thickness_pial'),
              ('outputspec.area_pial',
               'inputspec.lh_area_pial'), ('outputspec.volume',
                                           'inputspec.lh_volume'),
              ('outputspec.jacobian_white',
               'inputspec.lh_jacobian_white'), ('outputspec.sphere_reg',
                                                'inputspec.lh_sphere_reg')]),
            (ar3_lh_wf2, qcache_wf, [('outputspec.wg_pct_mgh',
                                      'inputspec.lh_wg_pct_mgh')]),
            (ar3_rh_wf1, qcache_wf,
             [('outputspec.thickness_pial', 'inputspec.rh_thickness_pial'),
              ('outputspec.area_pial',
               'inputspec.rh_area_pial'), ('outputspec.volume',
                                           'inputspec.rh_volume'),
              ('outputspec.jacobian_white',
               'inputspec.rh_jacobian_white'), ('outputspec.sphere_reg',
                                                'inputspec.rh_sphere_reg')]),
            (ar3_rh_wf2, qcache_wf, [('outputspec.wg_pct_mgh',
                                      'inputspec.rh_wg_pct_mgh')]),
        ])
        for source_file in source_inputs:
            ar3_wf.connect([(inputspec, source_subject, [('source_subject_dir',
                                                          'base_directory')]),
                            (source_subject, qcache_wf,
                             [(source_file,
                               'inputspec.source_' + source_file)])])
        # end qcache workflow

    # Add outputs to outputspec
    ar3_outputs = [
        'aseg', 'wmparc', 'wmparc_stats', 'aseg_stats', 'aparc_a2009s_aseg',
        'aparc_aseg', 'aseg_presurf_hypos', 'ribbon', 'rh_ribbon', 'lh_ribbon'
    ]
    for output in hemi_outputs1 + hemi_outputs2:
        for hemi in ('lh_', 'rh_'):
            ar3_outputs.append(hemi + output)
    if qcache:
        ar3_outputs.extend(qcache_outputs)

    ar3_outputs.extend(ba_outputs)

    outputspec = pe.Node(
        IdentityInterface(fields=ar3_outputs), name="outputspec")

    ar3_wf.connect([(apas_2_aseg, outputspec,
                     [('out_file', 'aseg')]), (wm_parcellation, outputspec,
                                               [('out_file', 'wmparc')]),
                    (wm_segstats, outputspec,
                     [('summary_file',
                       'wmparc_stats')]), (segstats, outputspec,
                                           [('summary_file', 'aseg_stats')]),
                    (aparc_2_aseg_2009, outputspec,
                     [('out_file',
                       'aparc_a2009s_aseg')]), (aparc_2_aseg, outputspec,
                                                [('out_file', 'aparc_aseg')]),
                    (volume_mask, outputspec,
                     [('out_ribbon', 'ribbon'), ('lh_ribbon', 'lh_ribbon'),
                      ('rh_ribbon', 'rh_ribbon')])])
    if fsvernum >= 6:
        ar3_wf.connect([(relabel_hypos, outputspec, [('out_file',
                                                      'aseg_presurf_hypos')])])

    for i, outputs in enumerate([hemi_outputs1, hemi_outputs2]):
        if i == 0:
            lhwf = ar3_lh_wf1
            rhwf = ar3_rh_wf1
        else:
            lhwf = ar3_lh_wf2
            rhwf = ar3_rh_wf2
        for output in outputs:
            ar3_wf.connect([(lhwf, outputspec, [('outputspec.' + output,
                                                 'lh_' + output)]),
                            (rhwf, outputspec, [('outputspec.' + output,
                                                 'rh_' + output)])])

    for output in ba_outputs:
        ar3_wf.connect([(ba_WF, outputspec, [('outputspec.' + output,
                                              output)])])

    if qcache:
        for output in qcache_outputs:
            ar3_wf.connect([(qcache_wf, outputspec, [('outputspec.' + output,
                                                      output)])])

    return ar3_wf, ar3_outputs



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..utils import MakeSurfaces


def test_MakeSurfaces_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        copy_inputs=dict(),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fix_mtl=dict(argstr='-fix_mtl', ),
        hemisphere=dict(
            argstr='%s',
            mandatory=True,
            position=-1,
        ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        in_T1=dict(argstr='-T1 %s', ),
        in_aseg=dict(argstr='-aseg %s', ),
        in_filled=dict(mandatory=True, ),
        in_label=dict(xor=['noaparc'], ),
        in_orig=dict(
            argstr='-orig %s',
            mandatory=True,
        ),
        in_white=dict(),
        in_wm=dict(mandatory=True, ),
        longitudinal=dict(argstr='-long', ),
        maximum=dict(argstr='-max %.1f', ),
        mgz=dict(argstr='-mgz', ),
        no_white=dict(argstr='-nowhite', ),
        noaparc=dict(
            argstr='-noaparc',
            xor=['in_label'],
        ),
        orig_pial=dict(
            argstr='-orig_pial %s',
            requires=['in_label'],
        ),
        orig_white=dict(argstr='-orig_white %s', ),
        subject_id=dict(
            argstr='%s',
            mandatory=True,
            position=-2,
            usedefault=True,
        ),
        subjects_dir=dict(),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
        white=dict(argstr='-white %s', ),
        white_only=dict(argstr='-whiteonly', ),
    )
    inputs = MakeSurfaces.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_MakeSurfaces_outputs():
    output_map = dict(
        out_area=dict(),
        out_cortex=dict(),
        out_curv=dict(),
        out_pial=dict(),
        out_thickness=dict(),
        out_white=dict(),
    )
    outputs = MakeSurfaces.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
from .preprocess import (create_susan_smooth, create_fsl_fs_preproc,
                         create_parallelfeat_preproc, create_featreg_preproc,
                         create_reg_workflow)
from .estimate import create_modelfit_workflow, create_fixed_effects_flow

# backwards compatibility
from ...rsfmri.fsl.resting import create_resting_preproc



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..utils import To3D


def test_To3D_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        assumemosaic=dict(argstr='-assume_dicom_mosaic', ),
        datatype=dict(argstr='-datum %s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        filetype=dict(argstr='-%s', ),
        funcparams=dict(argstr='-time:zt %s alt+z2', ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        in_folder=dict(
            argstr='%s/*.dcm',
            mandatory=True,
            position=-1,
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        out_file=dict(
            argstr='-prefix %s',
            name_source=['in_folder'],
            name_template='%s',
        ),
        outputtype=dict(),
        skipoutliers=dict(argstr='-skip_outliers', ),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
    )
    inputs = To3D.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_To3D_outputs():
    output_map = dict(out_file=dict(), )
    outputs = To3D.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..utils import Contrast


def test_Contrast_inputs():
    input_map = dict(
        annotation=dict(mandatory=True, ),
        args=dict(argstr='%s', ),
        copy_inputs=dict(),
        cortex=dict(mandatory=True, ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        hemisphere=dict(
            argstr='--%s-only',
            mandatory=True,
        ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        orig=dict(mandatory=True, ),
        rawavg=dict(mandatory=True, ),
        subject_id=dict(
            argstr='--s %s',
            mandatory=True,
            usedefault=True,
        ),
        subjects_dir=dict(),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
        thickness=dict(mandatory=True, ),
        white=dict(mandatory=True, ),
    )
    inputs = Contrast.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_Contrast_outputs():
    output_map = dict(
        out_contrast=dict(),
        out_log=dict(),
        out_stats=dict(),
    )
    outputs = Contrast.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..preprocess import ParseDICOMDir


def test_ParseDICOMDir_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        dicom_dir=dict(
            argstr='--d %s',
            mandatory=True,
        ),
        dicom_info_file=dict(
            argstr='--o %s',
            usedefault=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        sortbyrun=dict(argstr='--sortbyrun', ),
        subjects_dir=dict(),
        summarize=dict(argstr='--summarize', ),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
    )
    inputs = ParseDICOMDir.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_ParseDICOMDir_outputs():
    output_map = dict(dicom_info_file=dict(), )
    outputs = ParseDICOMDir.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""The fsl module provides classes for interfacing with the `FSL
<http://www.fmrib.ox.ac.uk/fsl/index.html>`_ command line tools.  This
was written to work with FSL version 4.1.4.
"""
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
from builtins import range, open

import os
import os.path as op
from warnings import warn

import numpy as np
from nibabel import load

from ...utils.filemanip import split_filename
from ..base import (TraitedSpec, File, InputMultiPath, OutputMultiPath,
                    Undefined, traits, isdefined)
from .base import FSLCommand, FSLCommandInputSpec, Info


class BETInputSpec(FSLCommandInputSpec):
    # We use position args here as list indices - so a negative number
    # will put something on the end
    in_file = File(
        exists=True,
        desc='input file to skull strip',
        argstr='%s',
        position=0,
        mandatory=True)
    out_file = File(
        desc='name of output skull stripped image',
        argstr='%s',
        position=1,
        genfile=True,
        hash_files=False)
    outline = traits.Bool(desc='create surface outline image', argstr='-o')
    mask = traits.Bool(desc='create binary mask image', argstr='-m')
    skull = traits.Bool(desc='create skull image', argstr='-s')
    no_output = traits.Bool(
        argstr='-n', desc="Don't generate segmented output")
    frac = traits.Float(
        desc='fractional intensity threshold', argstr='-f %.2f')
    vertical_gradient = traits.Float(
        argstr='-g %.2f',
        desc='vertical gradient in fractional intensity threshold (-1, 1)')
    radius = traits.Int(argstr='-r %d', units='mm', desc="head radius")
    center = traits.List(
        traits.Int,
        desc='center of gravity in voxels',
        argstr='-c %s',
        minlen=0,
        maxlen=3,
        units='voxels')
    threshold = traits.Bool(
        argstr='-t',
        desc="apply thresholding to segmented brain image and mask")
    mesh = traits.Bool(argstr='-e', desc="generate a vtk mesh brain surface")
    # the remaining 'options' are more like modes (mutually exclusive) that
    # FSL actually implements in a shell script wrapper around the bet binary.
    # for some combinations of them in specific order a call would not fail,
    # but in general using more than one of the following is clearly not
    # supported
    _xor_inputs = ('functional', 'reduce_bias', 'robust', 'padding',
                   'remove_eyes', 'surfaces', 't2_guided')
    robust = traits.Bool(
        desc='robust brain centre estimation (iterates BET several times)',
        argstr='-R',
        xor=_xor_inputs)
    padding = traits.Bool(
        desc=('improve BET if FOV is very small in Z (by temporarily padding '
              'end slices)'),
        argstr='-Z',
        xor=_xor_inputs)
    remove_eyes = traits.Bool(
        desc='eye & optic nerve cleanup (can be useful in SIENA)',
        argstr='-S',
        xor=_xor_inputs)
    surfaces = traits.Bool(
        desc=('run bet2 and then betsurf to get additional skull and scalp '
              'surfaces (includes registrations)'),
        argstr='-A',
        xor=_xor_inputs)
    t2_guided = File(
        desc='as with creating surfaces, when also feeding in '
        'non-brain-extracted T2 (includes registrations)',
        argstr='-A2 %s',
        xor=_xor_inputs)
    functional = traits.Bool(
        argstr='-F', xor=_xor_inputs, desc="apply to 4D fMRI data")
    reduce_bias = traits.Bool(
        argstr='-B', xor=_xor_inputs, desc="bias field and neck cleanup")


class BETOutputSpec(TraitedSpec):
    out_file = File(desc="path/name of skullstripped file (if generated)")
    mask_file = File(desc="path/name of binary brain mask (if generated)")
    outline_file = File(desc="path/name of outline file (if generated)")
    meshfile = File(desc="path/name of vtk mesh file (if generated)")
    inskull_mask_file = File(desc="path/name of inskull mask (if generated)")
    inskull_mesh_file = File(
        desc="path/name of inskull mesh outline (if generated)")
    outskull_mask_file = File(desc="path/name of outskull mask (if generated)")
    outskull_mesh_file = File(
        desc="path/name of outskull mesh outline (if generated)")
    outskin_mask_file = File(desc="path/name of outskin mask (if generated)")
    outskin_mesh_file = File(
        desc="path/name of outskin mesh outline (if generated)")
    skull_mask_file = File(desc="path/name of skull mask (if generated)")


class BET(FSLCommand):
    """Use FSL BET command for skull stripping.

    For complete details, see the `BET Documentation.
    <https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/BET/UserGuide>`_

    Examples
    --------
    >>> from nipype.interfaces import fsl
    >>> btr = fsl.BET()
    >>> btr.inputs.in_file = 'structural.nii'
    >>> btr.inputs.frac = 0.7
    >>> btr.inputs.out_file = 'brain_anat.nii'
    >>> btr.cmdline
    'bet structural.nii brain_anat.nii -f 0.70'
    >>> res = btr.run() # doctest: +SKIP

    """

    _cmd = 'bet'
    input_spec = BETInputSpec
    output_spec = BETOutputSpec

    def _run_interface(self, runtime):
        # The returncode is meaningless in BET.  So check the output
        # in stderr and if it's set, then update the returncode
        # accordingly.
        runtime = super(BET, self)._run_interface(runtime)
        if runtime.stderr:
            self.raise_exception(runtime)
        return runtime

    def _gen_outfilename(self):
        out_file = self.inputs.out_file
        if not isdefined(out_file) and isdefined(self.inputs.in_file):
            out_file = self._gen_fname(self.inputs.in_file, suffix='_brain')
        return os.path.abspath(out_file)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = self._gen_outfilename()
        if ((isdefined(self.inputs.mesh) and self.inputs.mesh)
                or (isdefined(self.inputs.surfaces) and self.inputs.surfaces)):
            outputs['meshfile'] = self._gen_fname(
                outputs['out_file'], suffix='_mesh.vtk', change_ext=False)
        if (isdefined(self.inputs.mask) and self.inputs.mask) or \
                (isdefined(self.inputs.reduce_bias) and
                 self.inputs.reduce_bias):
            outputs['mask_file'] = self._gen_fname(
                outputs['out_file'], suffix='_mask')
        if isdefined(self.inputs.outline) and self.inputs.outline:
            outputs['outline_file'] = self._gen_fname(
                outputs['out_file'], suffix='_overlay')
        if isdefined(self.inputs.surfaces) and self.inputs.surfaces:
            outputs['inskull_mask_file'] = self._gen_fname(
                outputs['out_file'], suffix='_inskull_mask')
            outputs['inskull_mesh_file'] = self._gen_fname(
                outputs['out_file'], suffix='_inskull_mesh')
            outputs['outskull_mask_file'] = self._gen_fname(
                outputs['out_file'], suffix='_outskull_mask')
            outputs['outskull_mesh_file'] = self._gen_fname(
                outputs['out_file'], suffix='_outskull_mesh')
            outputs['outskin_mask_file'] = self._gen_fname(
                outputs['out_file'], suffix='_outskin_mask')
            outputs['outskin_mesh_file'] = self._gen_fname(
                outputs['out_file'], suffix='_outskin_mesh')
            outputs['skull_mask_file'] = self._gen_fname(
                outputs['out_file'], suffix='_skull_mask')
        if isdefined(self.inputs.no_output) and self.inputs.no_output:
            outputs['out_file'] = Undefined
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            return self._gen_outfilename()
        return None


class FASTInputSpec(FSLCommandInputSpec):
    """ Defines inputs (trait classes) for FAST """
    in_files = InputMultiPath(
        File(exists=True),
        copyfile=False,
        desc='image, or multi-channel set of images, '
        'to be segmented',
        argstr='%s',
        position=-1,
        mandatory=True)
    out_basename = File(desc='base name of output files', argstr='-o %s')
    # ^^ uses in_file name as basename if none given
    number_classes = traits.Range(
        low=1, high=10, argstr='-n %d', desc='number of tissue-type classes')
    output_biasfield = traits.Bool(
        desc='output estimated bias field', argstr='-b')
    output_biascorrected = traits.Bool(
        desc='output restored image (bias-corrected image)', argstr='-B')
    img_type = traits.Enum(
        (1, 2, 3),
        desc='int specifying type of image: (1 = T1, 2 = T2, 3 = PD)',
        argstr='-t %d')
    bias_iters = traits.Range(
        low=1,
        high=10,
        argstr='-I %d',
        desc='number of main-loop iterations during '
        'bias-field removal')
    bias_lowpass = traits.Range(
        low=4,
        high=40,
        desc='bias field smoothing extent (FWHM) '
        'in mm',
        argstr='-l %d',
        units='mm')
    init_seg_smooth = traits.Range(
        low=0.0001,
        high=0.1,
        desc='initial segmentation spatial '
        'smoothness (during bias field '
        'estimation)',
        argstr='-f %.3f')
    segments = traits.Bool(
        desc='outputs a separate binary image for each '
        'tissue type',
        argstr='-g')
    init_transform = File(
        exists=True,
        desc='<standard2input.mat> initialise'
        ' using priors',
        argstr='-a %s')
    other_priors = InputMultiPath(
        File(exist=True),
        desc='alternative prior images',
        argstr='-A %s',
        minlen=3,
        maxlen=3)
    no_pve = traits.Bool(
        desc='turn off PVE (partial volume estimation)', argstr='--nopve')
    no_bias = traits.Bool(desc='do not remove bias field', argstr='-N')
    use_priors = traits.Bool(desc='use priors throughout', argstr='-P')
    # ^^ Must also set -a!, mutually inclusive?? No, conditional mandatory... need to figure out how to handle with traits.
    segment_iters = traits.Range(
        low=1,
        high=50,
        desc='number of segmentation-initialisation'
        ' iterations',
        argstr='-W %d')
    mixel_smooth = traits.Range(
        low=0.0,
        high=1.0,
        desc='spatial smoothness for mixeltype',
        argstr='-R %.2f')
    iters_afterbias = traits.Range(
        low=1,
        high=20,
        desc='number of main-loop iterations '
        'after bias-field removal',
        argstr='-O %d')
    hyper = traits.Range(
        low=0.0,
        high=1.0,
        desc='segmentation spatial smoothness',
        argstr='-H %.2f')
    verbose = traits.Bool(desc='switch on diagnostic messages', argstr='-v')
    manual_seg = File(
        exists=True, desc='Filename containing intensities', argstr='-s %s')
    probability_maps = traits.Bool(
        desc='outputs individual probability maps', argstr='-p')


class FASTOutputSpec(TraitedSpec):
    """Specify possible outputs from FAST"""
    tissue_class_map = File(
        exists=True,
        desc='path/name of binary segmented volume file'
        ' one val for each class  _seg')
    tissue_class_files = OutputMultiPath(
        File(
            desc=(
                'path/name of binary segmented volumes one file for each class  '
                '_seg_x')))
    restored_image = OutputMultiPath(
        File(
            desc=(
                'restored images (one for each input image) named according to '
                'the input images _restore')))

    mixeltype = File(desc="path/name of mixeltype volume file _mixeltype")

    partial_volume_map = File(desc='path/name of partial volume file _pveseg')
    partial_volume_files = OutputMultiPath(
        File(
            desc='path/name of partial volumes files one for each class, _pve_x'
        ))

    bias_field = OutputMultiPath(File(desc='Estimated bias field _bias'))
    probability_maps = OutputMultiPath(
        File(desc='filenames, one for each class, for each input, prob_x'))


class FAST(FSLCommand):
    """ Use FSL FAST for segmenting and bias correction.

    For complete details, see the `FAST Documentation.
    <https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FAST>`_

    Examples
    --------
    >>> from nipype.interfaces import fsl
    >>> fastr = fsl.FAST()
    >>> fastr.inputs.in_files = 'structural.nii'
    >>> fastr.inputs.out_basename = 'fast_'
    >>> fastr.cmdline
    'fast -o fast_ -S 1 structural.nii'
    >>> out = fastr.run()  # doctest: +SKIP

    """
    _cmd = 'fast'
    input_spec = FASTInputSpec
    output_spec = FASTOutputSpec

    def _format_arg(self, name, spec, value):
        # first do what should be done in general
        formatted = super(FAST, self)._format_arg(name, spec, value)
        if name == 'in_files':
            # FAST needs the -S parameter value to correspond to the number
            # of input images, otherwise it will ignore all but the first
            formatted = "-S %d %s" % (len(value), formatted)
        return formatted

    def _list_outputs(self):
        outputs = self.output_spec().get()
        if not isdefined(self.inputs.number_classes):
            nclasses = 3
        else:
            nclasses = self.inputs.number_classes
        # when using multichannel, results basename is based on last
        # input filename
        _gen_fname_opts = {}
        if isdefined(self.inputs.out_basename):
            _gen_fname_opts['basename'] = self.inputs.out_basename
            _gen_fname_opts['cwd'] = os.getcwd()
        else:
            _gen_fname_opts['basename'] = self.inputs.in_files[-1]
            _gen_fname_opts['cwd'], _, _ = split_filename(
                _gen_fname_opts['basename'])

        outputs['tissue_class_map'] = self._gen_fname(
            suffix='_seg', **_gen_fname_opts)
        if self.inputs.segments:
            outputs['tissue_class_files'] = []
            for i in range(nclasses):
                outputs['tissue_class_files'].append(
                    self._gen_fname(suffix='_seg_%d' % i, **_gen_fname_opts))
        if isdefined(self.inputs.output_biascorrected):
            outputs['restored_image'] = []
            if len(self.inputs.in_files) > 1:
                # for multi-image segmentation there is one corrected image
                # per input
                for val, f in enumerate(self.inputs.in_files):
                    # image numbering is 1-based
                    outputs['restored_image'].append(
                        self._gen_fname(
                            suffix='_restore_%d' % (val + 1),
                            **_gen_fname_opts))
            else:
                # single image segmentation has unnumbered output image
                outputs['restored_image'].append(
                    self._gen_fname(suffix='_restore', **_gen_fname_opts))

        outputs['mixeltype'] = self._gen_fname(
            suffix='_mixeltype', **_gen_fname_opts)
        if not self.inputs.no_pve:
            outputs['partial_volume_map'] = self._gen_fname(
                suffix='_pveseg', **_gen_fname_opts)
            outputs['partial_volume_files'] = []
            for i in range(nclasses):
                outputs['partial_volume_files'].append(
                    self._gen_fname(suffix='_pve_%d' % i, **_gen_fname_opts))
        if self.inputs.output_biasfield:
            outputs['bias_field'] = []
            if len(self.inputs.in_files) > 1:
                # for multi-image segmentation there is one bias field image
                # per input
                for val, f in enumerate(self.inputs.in_files):
                    # image numbering is 1-based
                    outputs['bias_field'].append(
                        self._gen_fname(
                            suffix='_bias_%d' % (val + 1), **_gen_fname_opts))
            else:
                # single image segmentation has unnumbered output image
                outputs['bias_field'].append(
                    self._gen_fname(suffix='_bias', **_gen_fname_opts))

        if self.inputs.probability_maps:
            outputs['probability_maps'] = []
            for i in range(nclasses):
                outputs['probability_maps'].append(
                    self._gen_fname(suffix='_prob_%d' % i, **_gen_fname_opts))
        return outputs


class FLIRTInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True,
        argstr='-in %s',
        mandatory=True,
        position=0,
        desc='input file')
    reference = File(
        exists=True,
        argstr='-ref %s',
        mandatory=True,
        position=1,
        desc='reference file')
    out_file = File(
        argstr='-out %s',
        desc='registered output file',
        name_source=['in_file'],
        name_template='%s_flirt',
        position=2,
        hash_files=False)
    out_matrix_file = File(
        argstr='-omat %s',
        name_source=['in_file'],
        keep_extension=True,
        name_template='%s_flirt.mat',
        desc='output affine matrix in 4x4 asciii format',
        position=3,
        hash_files=False)
    out_log = File(
        name_source=['in_file'],
        keep_extension=True,
        requires=['save_log'],
        name_template='%s_flirt.log',
        desc='output log')
    in_matrix_file = File(argstr='-init %s', desc='input 4x4 affine matrix')
    apply_xfm = traits.Bool(
        argstr='-applyxfm',
        desc=(
            'apply transformation supplied by in_matrix_file or uses_qform to'
            ' use the affine matrix stored in the reference header'))
    apply_isoxfm = traits.Float(
        argstr='-applyisoxfm %f',
        xor=['apply_xfm'],
        desc='as applyxfm but forces isotropic resampling')
    datatype = traits.Enum(
        'char',
        'short',
        'int',
        'float',
        'double',
        argstr='-datatype %s',
        desc='force output data type')
    cost = traits.Enum(
        'mutualinfo',
        'corratio',
        'normcorr',
        'normmi',
        'leastsq',
        'labeldiff',
        'bbr',
        argstr='-cost %s',
        desc='cost function')
    # XXX What is the difference between 'cost' and 'searchcost'?  Are
    # these both necessary or do they map to the same variable.
    cost_func = traits.Enum(
        'mutualinfo',
        'corratio',
        'normcorr',
        'normmi',
        'leastsq',
        'labeldiff',
        'bbr',
        argstr='-searchcost %s',
        desc='cost function')
    uses_qform = traits.Bool(
        argstr='-usesqform', desc='initialize using sform or qform')
    display_init = traits.Bool(
        argstr='-displayinit', desc='display initial matrix')
    angle_rep = traits.Enum(
        'quaternion',
        'euler',
        argstr='-anglerep %s',
        desc='representation of rotation angles')
    interp = traits.Enum(
        'trilinear',
        'nearestneighbour',
        'sinc',
        'spline',
        argstr='-interp %s',
        desc='final interpolation method used in reslicing')
    sinc_width = traits.Int(
        argstr='-sincwidth %d', units='voxels', desc='full-width in voxels')
    sinc_window = traits.Enum(
        'rectangular',
        'hanning',
        'blackman',
        argstr='-sincwindow %s',
        desc='sinc window')  # XXX better doc
    bins = traits.Int(argstr='-bins %d', desc='number of histogram bins')
    dof = traits.Int(
        argstr='-dof %d', desc='number of transform degrees of freedom')
    no_resample = traits.Bool(
        argstr='-noresample', desc='do not change input sampling')
    force_scaling = traits.Bool(
        argstr='-forcescaling', desc='force rescaling even for low-res images')
    min_sampling = traits.Float(
        argstr='-minsampling %f',
        units='mm',
        desc='set minimum voxel dimension for sampling')
    padding_size = traits.Int(
        argstr='-paddingsize %d',
        units='voxels',
        desc='for applyxfm: interpolates outside image '
        'by size')
    searchr_x = traits.List(
        traits.Int,
        minlen=2,
        maxlen=2,
        units='degrees',
        argstr='-searchrx %s',
        desc='search angles along x-axis, in degrees')
    searchr_y = traits.List(
        traits.Int,
        minlen=2,
        maxlen=2,
        units='degrees',
        argstr='-searchry %s',
        desc='search angles along y-axis, in degrees')
    searchr_z = traits.List(
        traits.Int,
        minlen=2,
        maxlen=2,
        units='degrees',
        argstr='-searchrz %s',
        desc='search angles along z-axis, in degrees')
    no_search = traits.Bool(
        argstr='-nosearch', desc='set all angular searches to ranges 0 to 0')
    coarse_search = traits.Int(
        argstr='-coarsesearch %d',
        units='degrees',
        desc='coarse search delta angle')
    fine_search = traits.Int(
        argstr='-finesearch %d',
        units='degrees',
        desc='fine search delta angle')
    schedule = File(
        exists=True, argstr='-schedule %s', desc='replaces default schedule')
    ref_weight = File(
        exists=True,
        argstr='-refweight %s',
        desc='File for reference weighting volume')
    in_weight = File(
        exists=True,
        argstr='-inweight %s',
        desc='File for input weighting volume')
    no_clamp = traits.Bool(
        argstr='-noclamp', desc='do not use intensity clamping')
    no_resample_blur = traits.Bool(
        argstr='-noresampblur', desc='do not use blurring on downsampling')
    rigid2D = traits.Bool(
        argstr='-2D', desc='use 2D rigid body mode - ignores dof')
    save_log = traits.Bool(desc='save to log file')
    verbose = traits.Int(argstr='-verbose %d', desc='verbose mode, 0 is least')
    bgvalue = traits.Float(
        0,
        argstr='-setbackground %f',
        desc=('use specified background value for points '
              'outside FOV'))

    # BBR options
    wm_seg = File(
        argstr='-wmseg %s',
        min_ver='5.0.0',
        desc='white matter segmentation volume needed by BBR cost function')
    wmcoords = File(
        argstr='-wmcoords %s',
        min_ver='5.0.0',
        desc='white matter boundary coordinates for BBR cost function')
    wmnorms = File(
        argstr='-wmnorms %s',
        min_ver='5.0.0',
        desc='white matter boundary normals for BBR cost function')
    fieldmap = File(
        argstr='-fieldmap %s',
        min_ver='5.0.0',
        desc=('fieldmap image in rads/s - must be already registered to the '
              'reference image'))
    fieldmapmask = File(
        argstr='-fieldmapmask %s',
        min_ver='5.0.0',
        desc='mask for fieldmap image')
    pedir = traits.Int(
        argstr='-pedir %d',
        min_ver='5.0.0',
        desc='phase encode direction of EPI - 1/2/3=x/y/z & -1/-2/-3=-x/-y/-z')
    echospacing = traits.Float(
        argstr='-echospacing %f',
        min_ver='5.0.0',
        desc='value of EPI echo spacing - units of seconds')
    bbrtype = traits.Enum(
        'signed',
        'global_abs',
        'local_abs',
        argstr='-bbrtype %s',
        min_ver='5.0.0',
        desc=('type of bbr cost function: signed [default], global_abs, '
              'local_abs'))
    bbrslope = traits.Float(
        argstr='-bbrslope %f', min_ver='5.0.0', desc='value of bbr slope')


class FLIRTOutputSpec(TraitedSpec):
    out_file = File(
        exists=True, desc='path/name of registered file (if generated)')
    out_matrix_file = File(
        exists=True,
        desc='path/name of calculated affine transform '
        '(if generated)')
    out_log = File(desc='path/name of output log (if generated)')


class FLIRT(FSLCommand):
    """Use FSL FLIRT for coregistration.

    For complete details, see the `FLIRT Documentation.
    <https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FLIRT>`_

    To print out the command line help, use:
        fsl.FLIRT().inputs_help()

    Examples
    --------
    >>> from nipype.interfaces import fsl
    >>> from nipype.testing import example_data
    >>> flt = fsl.FLIRT(bins=640, cost_func='mutualinfo')
    >>> flt.inputs.in_file = 'structural.nii'
    >>> flt.inputs.reference = 'mni.nii'
    >>> flt.inputs.output_type = "NIFTI_GZ"
    >>> flt.cmdline # doctest: +ELLIPSIS
    'flirt -in structural.nii -ref mni.nii -out structural_flirt.nii.gz -omat structural_flirt.mat -bins 640 -searchcost mutualinfo'
    >>> res = flt.run() #doctest: +SKIP

    """
    _cmd = 'flirt'
    input_spec = FLIRTInputSpec
    output_spec = FLIRTOutputSpec
    _log_written = False

    def aggregate_outputs(self, runtime=None, needed_outputs=None):
        outputs = super(FLIRT, self).aggregate_outputs(
            runtime=runtime, needed_outputs=needed_outputs)
        if self.inputs.save_log and not self._log_written:
            with open(outputs.out_log, "a") as text_file:
                text_file.write(runtime.stdout + '\n')
            self._log_written = True
        return outputs

    def _parse_inputs(self, skip=None):
        if skip is None:
            skip = []
        if self.inputs.save_log and not self.inputs.verbose:
            self.inputs.verbose = 1
        if self.inputs.apply_xfm and not (self.inputs.in_matrix_file
                                          or self.inputs.uses_qform):
            raise RuntimeError('Argument apply_xfm requires in_matrix_file or '
                               'uses_qform arguments to run')
        skip.append('save_log')
        return super(FLIRT, self)._parse_inputs(skip=skip)


class ApplyXFMInputSpec(FLIRTInputSpec):
    apply_xfm = traits.Bool(
        True,
        argstr='-applyxfm',
        desc=(
            'apply transformation supplied by in_matrix_file or uses_qform to'
            ' use the affine matrix stored in the reference header'),
        usedefault=True)


class ApplyXFM(FLIRT):
    """Currently just a light wrapper around FLIRT,
    with no modifications

    ApplyXFM is used to apply an existing tranform to an image


    Examples
    --------

    >>> import nipype.interfaces.fsl as fsl
    >>> from nipype.testing import example_data
    >>> applyxfm = fsl.preprocess.ApplyXFM()
    >>> applyxfm.inputs.in_file = example_data('structural.nii')
    >>> applyxfm.inputs.in_matrix_file = example_data('trans.mat')
    >>> applyxfm.inputs.out_file = 'newfile.nii'
    >>> applyxfm.inputs.reference = example_data('mni.nii')
    >>> applyxfm.inputs.apply_xfm = True
    >>> result = applyxfm.run() # doctest: +SKIP

    """
    input_spec = ApplyXFMInputSpec


class MCFLIRTInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True,
        position=0,
        argstr="-in %s",
        mandatory=True,
        desc="timeseries to motion-correct")
    out_file = File(
        argstr='-out %s', genfile=True, desc="file to write", hash_files=False)
    cost = traits.Enum(
        'mutualinfo',
        'woods',
        'corratio',
        'normcorr',
        'normmi',
        'leastsquares',
        argstr='-cost %s',
        desc="cost function to optimize")
    bins = traits.Int(argstr='-bins %d', desc="number of histogram bins")
    dof = traits.Int(
        argstr='-dof %d', desc="degrees of freedom for the transformation")
    ref_vol = traits.Int(argstr='-refvol %d', desc="volume to align frames to")
    scaling = traits.Float(
        argstr='-scaling %.2f', desc="scaling factor to use")
    smooth = traits.Float(
        argstr='-smooth %.2f', desc="smoothing factor for the cost function")
    rotation = traits.Int(
        argstr='-rotation %d', desc="scaling factor for rotation tolerances")
    stages = traits.Int(
        argstr='-stages %d',
        desc="stages (if 4, perform final search with sinc interpolation")
    init = File(
        exists=True, argstr='-init %s', desc="inital transformation matrix")
    interpolation = traits.Enum(
        "spline",
        "nn",
        "sinc",
        argstr="-%s_final",
        desc="interpolation method for transformation")
    use_gradient = traits.Bool(
        argstr='-gdt', desc="run search on gradient images")
    use_contour = traits.Bool(
        argstr='-edge', desc="run search on contour images")
    mean_vol = traits.Bool(argstr='-meanvol', desc="register to mean volume")
    stats_imgs = traits.Bool(
        argstr='-stats', desc="produce variance and std. dev. images")
    save_mats = traits.Bool(
        argstr='-mats', desc="save transformation matrices")
    save_plots = traits.Bool(
        argstr='-plots', desc="save transformation parameters")
    save_rms = traits.Bool(
        argstr='-rmsabs -rmsrel', desc="save rms displacement parameters")
    ref_file = File(
        exists=True,
        argstr='-reffile %s',
        desc="target image for motion correction")


class MCFLIRTOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="motion-corrected timeseries")
    variance_img = File(exists=True, desc="variance image")
    std_img = File(exists=True, desc="standard deviation image")
    mean_img = File(
        exists=True, desc="mean timeseries image (if mean_vol=True)")
    par_file = File(exists=True, desc="text-file with motion parameters")
    mat_file = OutputMultiPath(
        File(exists=True), desc="transformation matrices")
    rms_files = OutputMultiPath(
        File(exists=True),
        desc="absolute and relative displacement parameters")


class MCFLIRT(FSLCommand):
    """Use FSL MCFLIRT to do within-modality motion correction.

    For complete details, see the `MCFLIRT Documentation.
    <https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/MCFLIRT>`_

    Examples
    --------
    >>> from nipype.interfaces import fsl
    >>> mcflt = fsl.MCFLIRT()
    >>> mcflt.inputs.in_file = 'functional.nii'
    >>> mcflt.inputs.cost = 'mutualinfo'
    >>> mcflt.inputs.out_file = 'moco.nii'
    >>> mcflt.cmdline
    'mcflirt -in functional.nii -cost mutualinfo -out moco.nii'
    >>> res = mcflt.run()  # doctest: +SKIP

    """
    _cmd = 'mcflirt'
    input_spec = MCFLIRTInputSpec
    output_spec = MCFLIRTOutputSpec

    def _format_arg(self, name, spec, value):
        if name == "interpolation":
            if value == "trilinear":
                return ""
            else:
                return spec.argstr % value
        return super(MCFLIRT, self)._format_arg(name, spec, value)

    def _list_outputs(self):
        cwd = os.getcwd()
        outputs = self._outputs().get()

        outputs['out_file'] = self._gen_outfilename()

        if isdefined(self.inputs.stats_imgs) and self.inputs.stats_imgs:
            outputs['variance_img'] = self._gen_fname(
                outputs['out_file'] + '_variance.ext', cwd=cwd)
            outputs['std_img'] = self._gen_fname(
                outputs['out_file'] + '_sigma.ext', cwd=cwd)

        # The mean image created if -stats option is specified ('meanvol')
        # is missing the top and bottom slices. Therefore we only expose the
        # mean image created by -meanvol option ('mean_reg') which isn't
        # corrupted.
        # Note that the same problem holds for the std and variance image.

        if isdefined(self.inputs.mean_vol) and self.inputs.mean_vol:
            outputs['mean_img'] = self._gen_fname(
                outputs['out_file'] + '_mean_reg.ext', cwd=cwd)

        if isdefined(self.inputs.save_mats) and self.inputs.save_mats:
            _, filename = os.path.split(outputs['out_file'])
            matpathname = os.path.join(cwd, filename + '.mat')
            _, _, _, timepoints = load(self.inputs.in_file).shape
            outputs['mat_file'] = []
            for t in range(timepoints):
                outputs['mat_file'].append(
                    os.path.join(matpathname, 'MAT_%04d' % t))
        if isdefined(self.inputs.save_plots) and self.inputs.save_plots:
            # Note - if e.g. out_file has .nii.gz, you get .nii.gz.par,
            # which is what mcflirt does!
            outputs['par_file'] = outputs['out_file'] + '.par'
        if isdefined(self.inputs.save_rms) and self.inputs.save_rms:
            outfile = outputs['out_file']
            outputs['rms_files'] = [outfile + '_abs.rms', outfile + '_rel.rms']
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            return self._gen_outfilename()
        return None

    def _gen_outfilename(self):
        out_file = self.inputs.out_file
        if isdefined(out_file):
            out_file = os.path.realpath(out_file)
        if not isdefined(out_file) and isdefined(self.inputs.in_file):
            out_file = self._gen_fname(self.inputs.in_file, suffix='_mcf')
        return os.path.abspath(out_file)


class FNIRTInputSpec(FSLCommandInputSpec):
    ref_file = File(
        exists=True,
        argstr='--ref=%s',
        mandatory=True,
        desc='name of reference image')
    in_file = File(
        exists=True,
        argstr='--in=%s',
        mandatory=True,
        desc='name of input image')
    affine_file = File(
        exists=True,
        argstr='--aff=%s',
        desc='name of file containing affine transform')
    inwarp_file = File(
        exists=True,
        argstr='--inwarp=%s',
        desc='name of file containing initial non-linear warps')
    in_intensitymap_file = traits.List(
        File(exists=True),
        argstr='--intin=%s',
        copyfile=False,
        minlen=1,
        maxlen=2,
        desc=('name of file/files containing '
              'initial intensity mapping '
              'usually generated by previous '
              'fnirt run'))
    fieldcoeff_file = traits.Either(
        traits.Bool,
        File,
        argstr='--cout=%s',
        desc='name of output file with field coefficients or true')
    warped_file = File(
        argstr='--iout=%s',
        desc='name of output image',
        genfile=True,
        hash_files=False)
    field_file = traits.Either(
        traits.Bool,
        File,
        argstr='--fout=%s',
        desc='name of output file with field or true',
        hash_files=False)
    jacobian_file = traits.Either(
        traits.Bool,
        File,
        argstr='--jout=%s',
        desc=('name of file for writing out the '
              'Jacobian of the field (for '
              'diagnostic or VBM purposes)'),
        hash_files=False)
    modulatedref_file = traits.Either(
        traits.Bool,
        File,
        argstr='--refout=%s',
        desc=('name of file for writing out '
              'intensity modulated --ref (for '
              'diagnostic purposes)'),
        hash_files=False)
    out_intensitymap_file = traits.Either(
        traits.Bool,
        File,
        argstr='--intout=%s',
        desc=('name of files for writing '
              'information pertaining to '
              'intensity mapping'),
        hash_files=False)
    log_file = File(
        argstr='--logout=%s',
        desc='Name of log-file',
        genfile=True,
        hash_files=False)
    config_file = traits.Either(
        traits.Enum("T1_2_MNI152_2mm", "FA_2_FMRIB58_1mm"),
        File(exists=True),
        argstr='--config=%s',
        desc='Name of config file specifying command line arguments')
    refmask_file = File(
        exists=True,
        argstr='--refmask=%s',
        desc='name of file with mask in reference space')
    inmask_file = File(
        exists=True,
        argstr='--inmask=%s',
        desc='name of file with mask in input image space')
    skip_refmask = traits.Bool(
        argstr='--applyrefmask=0',
        xor=['apply_refmask'],
        desc='Skip specified refmask if set, default false')
    skip_inmask = traits.Bool(
        argstr='--applyinmask=0',
        xor=['apply_inmask'],
        desc='skip specified inmask if set, default false')
    apply_refmask = traits.List(
        traits.Enum(0, 1),
        argstr='--applyrefmask=%s',
        xor=['skip_refmask'],
        desc=('list of iterations to use reference mask on (1 to use, 0 to '
              'skip)'),
        sep=",")
    apply_inmask = traits.List(
        traits.Enum(0, 1),
        argstr='--applyinmask=%s',
        xor=['skip_inmask'],
        desc='list of iterations to use input mask on (1 to use, 0 to skip)',
        sep=",")
    skip_implicit_ref_masking = traits.Bool(
        argstr='--imprefm=0',
        desc=('skip implicit masking  based on value in --ref image. '
              'Default = 0'))
    skip_implicit_in_masking = traits.Bool(
        argstr='--impinm=0',
        desc=('skip implicit masking  based on value in --in image. '
              'Default = 0'))
    refmask_val = traits.Float(
        argstr='--imprefval=%f',
        desc='Value to mask out in --ref image. Default =0.0')
    inmask_val = traits.Float(
        argstr='--impinval=%f',
        desc='Value to mask out in --in image. Default =0.0')
    max_nonlin_iter = traits.List(
        traits.Int,
        argstr='--miter=%s',
        desc='Max # of non-linear iterations list, default [5, 5, 5, 5]',
        sep=",")
    subsampling_scheme = traits.List(
        traits.Int,
        argstr='--subsamp=%s',
        desc='sub-sampling scheme, list, default [4, 2, 1, 1]',
        sep=",")
    warp_resolution = traits.Tuple(
        traits.Int,
        traits.Int,
        traits.Int,
        argstr='--warpres=%d,%d,%d',
        desc=('(approximate) resolution (in mm) of warp basis in x-, y- and '
              'z-direction, default 10, 10, 10'))
    spline_order = traits.Int(
        argstr='--splineorder=%d',
        desc='Order of spline, 2->Qadratic spline, 3->Cubic spline. Default=3')
    in_fwhm = traits.List(
        traits.Int,
        argstr='--infwhm=%s',
        desc=('FWHM (in mm) of gaussian smoothing kernel for input volume, '
              'default [6, 4, 2, 2]'),
        sep=",")
    ref_fwhm = traits.List(
        traits.Int,
        argstr='--reffwhm=%s',
        desc=('FWHM (in mm) of gaussian smoothing kernel for ref volume, '
              'default [4, 2, 0, 0]'),
        sep=",")
    regularization_model = traits.Enum(
        'membrane_energy',
        'bending_energy',
        argstr='--regmod=%s',
        desc=('Model for regularisation of warp-field [membrane_energy '
              'bending_energy], default bending_energy'))
    regularization_lambda = traits.List(
        traits.Float,
        argstr='--lambda=%s',
        desc=('Weight of regularisation, default depending on --ssqlambda and '
              '--regmod switches. See user documetation.'),
        sep=",")
    skip_lambda_ssq = traits.Bool(
        argstr='--ssqlambda=0',
        desc='If true, lambda is not weighted by current ssq, default false')
    jacobian_range = traits.Tuple(
        traits.Float,
        traits.Float,
        argstr='--jacrange=%f,%f',
        desc='Allowed range of Jacobian determinants, default 0.01, 100.0')
    derive_from_ref = traits.Bool(
        argstr='--refderiv',
        desc=('If true, ref image is used to calculate derivatives. '
              'Default false'))
    intensity_mapping_model = traits.Enum(
        'none',
        'global_linear',
        'global_non_linear',
        'local_linear',
        'global_non_linear_with_bias',
        'local_non_linear',
        argstr='--intmod=%s',
        desc='Model for intensity-mapping')
    intensity_mapping_order = traits.Int(
        argstr='--intorder=%d',
        desc='Order of poynomial for mapping intensities, default 5')
    biasfield_resolution = traits.Tuple(
        traits.Int,
        traits.Int,
        traits.Int,
        argstr='--biasres=%d,%d,%d',
        desc=('Resolution (in mm) of bias-field modelling local intensities, '
              'default 50, 50, 50'))
    bias_regularization_lambda = traits.Float(
        argstr='--biaslambda=%f',
        desc='Weight of regularisation for bias-field, default 10000')
    skip_intensity_mapping = traits.Bool(
        argstr='--estint=0',
        xor=['apply_intensity_mapping'],
        desc='Skip estimate intensity-mapping default false')
    apply_intensity_mapping = traits.List(
        traits.Enum(0, 1),
        argstr='--estint=%s',
        xor=['skip_intensity_mapping'],
        desc=('List of subsampling levels to apply intensity mapping for '
              '(0 to skip, 1 to apply)'),
        sep=",")
    hessian_precision = traits.Enum(
        'double',
        'float',
        argstr='--numprec=%s',
        desc=('Precision for representing Hessian, double or float. '
              'Default double'))


class FNIRTOutputSpec(TraitedSpec):
    fieldcoeff_file = File(exists=True, desc='file with field coefficients')
    warped_file = File(exists=True, desc='warped image')
    field_file = File(desc='file with warp field')
    jacobian_file = File(desc='file containing Jacobian of the field')
    modulatedref_file = File(desc='file containing intensity modulated --ref')
    out_intensitymap_file = traits.List(
        File,
        minlen=2,
        maxlen=2,
        desc='files containing info pertaining to intensity mapping')
    log_file = File(desc='Name of log-file')


class FNIRT(FSLCommand):
    """Use FSL FNIRT for non-linear registration.

    For complete details, see the `FNIRT Documentation.
    <https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FNIRT>`_

    Examples
    --------
    >>> from nipype.interfaces import fsl
    >>> from nipype.testing import example_data
    >>> fnt = fsl.FNIRT(affine_file=example_data('trans.mat'))
    >>> res = fnt.run(ref_file=example_data('mni.nii', in_file=example_data('structural.nii')) #doctest: +SKIP

    T1 -> Mni153

    >>> from nipype.interfaces import fsl
    >>> fnirt_mprage = fsl.FNIRT()
    >>> fnirt_mprage.inputs.in_fwhm = [8, 4, 2, 2]
    >>> fnirt_mprage.inputs.subsampling_scheme = [4, 2, 1, 1]

    Specify the resolution of the warps

    >>> fnirt_mprage.inputs.warp_resolution = (6, 6, 6)
    >>> res = fnirt_mprage.run(in_file='structural.nii', ref_file='mni.nii', warped_file='warped.nii', fieldcoeff_file='fieldcoeff.nii')#doctest: +SKIP

    We can check the command line and confirm that it's what we expect.

    >>> fnirt_mprage.cmdline  #doctest: +SKIP
    'fnirt --cout=fieldcoeff.nii --in=structural.nii --infwhm=8,4,2,2 --ref=mni.nii --subsamp=4,2,1,1 --warpres=6,6,6 --iout=warped.nii'

    """

    _cmd = 'fnirt'
    input_spec = FNIRTInputSpec
    output_spec = FNIRTOutputSpec

    filemap = {
        'warped_file': 'warped',
        'field_file': 'field',
        'jacobian_file': 'field_jacobian',
        'modulatedref_file': 'modulated',
        'out_intensitymap_file': 'intmap',
        'log_file': 'log.txt',
        'fieldcoeff_file': 'fieldwarp'
    }

    def _list_outputs(self):
        outputs = self.output_spec().get()
        for key, suffix in list(self.filemap.items()):
            inval = getattr(self.inputs, key)
            change_ext = True
            if key in ['warped_file', 'log_file']:
                if suffix.endswith('.txt'):
                    change_ext = False
                if isdefined(inval):
                    outputs[key] = inval
                else:
                    outputs[key] = self._gen_fname(
                        self.inputs.in_file,
                        suffix='_' + suffix,
                        change_ext=change_ext)
            elif isdefined(inval):
                if isinstance(inval, bool):
                    if inval:
                        outputs[key] = self._gen_fname(
                            self.inputs.in_file,
                            suffix='_' + suffix,
                            change_ext=change_ext)
                else:
                    outputs[key] = os.path.abspath(inval)

            if key == 'out_intensitymap_file' and isdefined(outputs[key]):
                basename = FNIRT.intensitymap_file_basename(outputs[key])
                outputs[key] = [
                    outputs[key],
                    '%s.txt' % basename,
                ]
        return outputs

    def _format_arg(self, name, spec, value):
        if name in ('in_intensitymap_file', 'out_intensitymap_file'):
            if name == 'out_intensitymap_file':
                value = self._list_outputs()[name]
            value = [FNIRT.intensitymap_file_basename(v) for v in value]
            assert len(set(value)) == 1, (
                'Found different basenames for {}: {}'.format(name, value))
            return spec.argstr % value[0]
        if name in list(self.filemap.keys()):
            return spec.argstr % self._list_outputs()[name]
        return super(FNIRT, self)._format_arg(name, spec, value)

    def _gen_filename(self, name):
        if name in ['warped_file', 'log_file']:
            return self._list_outputs()[name]
        return None

    def write_config(self, configfile):
        """Writes out currently set options to specified config file

        XX TODO : need to figure out how the config file is written

        Parameters
        ----------
        configfile : /path/to/configfile
        """
        try:
            fid = open(configfile, 'w+')
        except IOError:
            print('unable to create config_file %s' % (configfile))

        for item in list(self.inputs.get().items()):
            fid.write('%s\n' % (item))
        fid.close()

    @classmethod
    def intensitymap_file_basename(cls, f):
        """Removes valid intensitymap extensions from `f`, returning a basename
        that can refer to both intensitymap files.
        """
        for ext in list(Info.ftypes.values()) + ['.txt']:
            if f.endswith(ext):
                return f[:-len(ext)]
        # TODO consider warning for this case
        return f


class ApplyWarpInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True,
        argstr='--in=%s',
        mandatory=True,
        position=0,
        desc='image to be warped')
    out_file = File(
        argstr='--out=%s',
        genfile=True,
        position=2,
        desc='output filename',
        hash_files=False)
    ref_file = File(
        exists=True,
        argstr='--ref=%s',
        mandatory=True,
        position=1,
        desc='reference image')
    field_file = File(
        exists=True, argstr='--warp=%s', desc='file containing warp field')
    abswarp = traits.Bool(
        argstr='--abs',
        xor=['relwarp'],
        desc="treat warp field as absolute: x' = w(x)")
    relwarp = traits.Bool(
        argstr='--rel',
        xor=['abswarp'],
        position=-1,
        desc="treat warp field as relative: x' = x + w(x)")
    datatype = traits.Enum(
        'char',
        'short',
        'int',
        'float',
        'double',
        argstr='--datatype=%s',
        desc='Force output data type [char short int float double].')
    supersample = traits.Bool(
        argstr='--super',
        desc='intermediary supersampling of output, default is off')
    superlevel = traits.Either(
        traits.Enum('a'),
        traits.Int,
        argstr='--superlevel=%s',
        desc=("level of intermediary supersampling, a for 'automatic' or "
              "integer level. Default = 2"))
    premat = File(
        exists=True,
        argstr='--premat=%s',
        desc='filename for pre-transform (affine matrix)')
    postmat = File(
        exists=True,
        argstr='--postmat=%s',
        desc='filename for post-transform (affine matrix)')
    mask_file = File(
        exists=True,
        argstr='--mask=%s',
        desc='filename for mask image (in reference space)')
    interp = traits.Enum(
        'nn',
        'trilinear',
        'sinc',
        'spline',
        argstr='--interp=%s',
        position=-2,
        desc='interpolation method')


class ApplyWarpOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='Warped output file')


class ApplyWarp(FSLCommand):
    """Use FSL's applywarp to apply the results of a FNIRT registration

    Examples
    --------
    >>> from nipype.interfaces import fsl
    >>> from nipype.testing import example_data
    >>> aw = fsl.ApplyWarp()
    >>> aw.inputs.in_file = example_data('structural.nii')
    >>> aw.inputs.ref_file = example_data('mni.nii')
    >>> aw.inputs.field_file = 'my_coefficients_filed.nii' #doctest: +SKIP
    >>> res = aw.run() #doctest: +SKIP


    """

    _cmd = 'applywarp'
    input_spec = ApplyWarpInputSpec
    output_spec = ApplyWarpOutputSpec

    def _format_arg(self, name, spec, value):
        if name == 'superlevel':
            return spec.argstr % str(value)
        return super(ApplyWarp, self)._format_arg(name, spec, value)

    def _list_outputs(self):
        outputs = self._outputs().get()
        if not isdefined(self.inputs.out_file):
            outputs['out_file'] = self._gen_fname(
                self.inputs.in_file, suffix='_warp')
        else:
            outputs['out_file'] = os.path.abspath(self.inputs.out_file)
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            return self._list_outputs()[name]
        return None


class SliceTimerInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True,
        argstr='--in=%s',
        mandatory=True,
        position=0,
        desc='filename of input timeseries')
    out_file = File(
        argstr='--out=%s',
        genfile=True,
        desc='filename of output timeseries',
        hash_files=False)
    index_dir = traits.Bool(
        argstr='--down', desc='slice indexing from top to bottom')
    time_repetition = traits.Float(
        argstr='--repeat=%f', desc='Specify TR of data - default is 3s')
    slice_direction = traits.Enum(
        1,
        2,
        3,
        argstr='--direction=%d',
        desc='direction of slice acquisition (x=1, y=2, z=3) - default is z')
    interleaved = traits.Bool(
        argstr='--odd', desc='use interleaved acquisition')
    custom_timings = File(
        exists=True,
        argstr='--tcustom=%s',
        desc=('slice timings, in fractions of TR, range 0:1 (default is 0.5 = '
              'no shift)'))
    global_shift = traits.Float(
        argstr='--tglobal',
        desc='shift in fraction of TR, range 0:1 (default is 0.5 = no shift)')
    custom_order = File(
        exists=True,
        argstr='--ocustom=%s',
        desc=('filename of single-column custom interleave order file (first '
              'slice is referred to as 1 not 0)'))


class SliceTimerOutputSpec(TraitedSpec):
    slice_time_corrected_file = File(
        exists=True, desc='slice time corrected file')


class SliceTimer(FSLCommand):
    """ use FSL slicetimer to perform slice timing correction.

    Examples
    --------
    >>> from nipype.interfaces import fsl
    >>> from nipype.testing import example_data
    >>> st = fsl.SliceTimer()
    >>> st.inputs.in_file = example_data('functional.nii')
    >>> st.inputs.interleaved = True
    >>> result = st.run() #doctest: +SKIP

    """

    _cmd = 'slicetimer'
    input_spec = SliceTimerInputSpec
    output_spec = SliceTimerOutputSpec

    def _list_outputs(self):
        outputs = self._outputs().get()
        out_file = self.inputs.out_file
        if not isdefined(out_file):
            out_file = self._gen_fname(self.inputs.in_file, suffix='_st')
        outputs['slice_time_corrected_file'] = os.path.abspath(out_file)
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            return self._list_outputs()['slice_time_corrected_file']
        return None


class SUSANInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True,
        argstr='%s',
        mandatory=True,
        position=1,
        desc='filename of input timeseries')
    brightness_threshold = traits.Float(
        argstr='%.10f',
        position=2,
        mandatory=True,
        desc=('brightness threshold and should be greater than noise level '
              'and less than contrast of edges to be preserved.'))
    fwhm = traits.Float(
        argstr='%.10f',
        position=3,
        mandatory=True,
        desc='fwhm of smoothing, in mm, gets converted using sqrt(8*log(2))')
    dimension = traits.Enum(
        3,
        2,
        argstr='%d',
        position=4,
        usedefault=True,
        desc='within-plane (2) or fully 3D (3)')
    use_median = traits.Enum(
        1,
        0,
        argstr='%d',
        position=5,
        usedefault=True,
        desc=('whether to use a local median filter in the cases where '
              'single-point noise is detected'))
    usans = traits.List(
        traits.Tuple(File(exists=True), traits.Float),
        maxlen=2,
        argstr='',
        position=6,
        usedefault=True,
        desc='determines whether the smoothing area (USAN) is to be '
        'found from secondary images (0, 1 or 2). A negative '
        'value for any brightness threshold will auto-set the '
        'threshold at 10% of the robust range')
    out_file = File(
        argstr='%s',
        position=-1,
        genfile=True,
        desc='output file name',
        hash_files=False)


class SUSANOutputSpec(TraitedSpec):
    smoothed_file = File(exists=True, desc='smoothed output file')


class SUSAN(FSLCommand):
    """ use FSL SUSAN to perform smoothing

    For complete details, see the `SUSAN Documentation.
    <https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/SUSAN>`_

    Examples
    --------

    >>> from nipype.interfaces import fsl
    >>> from nipype.testing import example_data
    >>> anatfile  # doctest: +SKIP
    anatomical.nii  # doctest: +SKIP
    >>> sus = fsl.SUSAN()
    >>> sus.inputs.in_file = example_data('structural.nii')
    >>> sus.inputs.brightness_threshold = 2000.0
    >>> sus.inputs.fwhm = 8.0
    >>> result = sus.run()  # doctest: +SKIP
    """

    _cmd = 'susan'
    input_spec = SUSANInputSpec
    output_spec = SUSANOutputSpec

    def _format_arg(self, name, spec, value):
        if name == 'fwhm':
            return spec.argstr % (float(value) / np.sqrt(8 * np.log(2)))
        if name == 'usans':
            if not value:
                return '0'
            arglist = [str(len(value))]
            for filename, thresh in value:
                arglist.extend([filename, '%.10f' % thresh])
            return ' '.join(arglist)
        return super(SUSAN, self)._format_arg(name, spec, value)

    def _list_outputs(self):
        outputs = self._outputs().get()
        out_file = self.inputs.out_file
        if not isdefined(out_file):
            out_file = self._gen_fname(self.inputs.in_file, suffix='_smooth')
        outputs['smoothed_file'] = os.path.abspath(out_file)
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            return self._list_outputs()['smoothed_file']
        return None


class FUGUEInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True, argstr='--in=%s', desc='filename of input volume')
    shift_in_file = File(
        exists=True,
        argstr='--loadshift=%s',
        desc='filename for reading pixel shift volume')
    phasemap_in_file = File(
        exists=True,
        argstr='--phasemap=%s',
        desc='filename for input phase image')
    fmap_in_file = File(
        exists=True,
        argstr='--loadfmap=%s',
        desc='filename for loading fieldmap (rad/s)')
    unwarped_file = File(
        argstr='--unwarp=%s',
        desc='apply unwarping and save as filename',
        xor=['warped_file'],
        requires=['in_file'])
    warped_file = File(
        argstr='--warp=%s',
        desc='apply forward warping and save as filename',
        xor=['unwarped_file'],
        requires=['in_file'])
    forward_warping = traits.Bool(
        False,
        usedefault=True,
        desc='apply forward warping instead of unwarping')
    dwell_to_asym_ratio = traits.Float(
        argstr='--dwelltoasym=%.10f', desc='set the dwell to asym time ratio')
    dwell_time = traits.Float(
        argstr='--dwell=%.10f',
        desc=('set the EPI dwell time per phase-encode line - same as echo '
              'spacing - (sec)'))
    asym_se_time = traits.Float(
        argstr='--asym=%.10f',
        desc='set the fieldmap asymmetric spin echo time (sec)')
    median_2dfilter = traits.Bool(
        argstr='--median', desc='apply 2D median filtering')
    despike_2dfilter = traits.Bool(
        argstr='--despike', desc='apply a 2D de-spiking filter')
    no_gap_fill = traits.Bool(
        argstr='--nofill',
        desc='do not apply gap-filling measure to the fieldmap')
    no_extend = traits.Bool(
        argstr='--noextend',
        desc='do not apply rigid-body extrapolation to the fieldmap')
    smooth2d = traits.Float(
        argstr='--smooth2=%.2f',
        desc='apply 2D Gaussian smoothing of sigma N (in mm)')
    smooth3d = traits.Float(
        argstr='--smooth3=%.2f',
        desc='apply 3D Gaussian smoothing of sigma N (in mm)')
    poly_order = traits.Int(
        argstr='--poly=%d', desc='apply polynomial fitting of order N')
    fourier_order = traits.Int(
        argstr='--fourier=%d',
        desc='apply Fourier (sinusoidal) fitting of order N')
    pava = traits.Bool(
        argstr='--pava', desc='apply monotonic enforcement via PAVA')
    despike_threshold = traits.Float(
        argstr='--despikethreshold=%s',
        desc='specify the threshold for de-spiking (default=3.0)')
    unwarp_direction = traits.Enum(
        'x',
        'y',
        'z',
        'x-',
        'y-',
        'z-',
        argstr='--unwarpdir=%s',
        desc='specifies direction of warping (default y)')
    phase_conjugate = traits.Bool(
        argstr='--phaseconj', desc='apply phase conjugate method of unwarping')
    icorr = traits.Bool(
        argstr='--icorr',
        requires=['shift_in_file'],
        desc=('apply intensity correction to unwarping (pixel shift method '
              'only)'))
    icorr_only = traits.Bool(
        argstr='--icorronly',
        requires=['unwarped_file'],
        desc='apply intensity correction only')
    mask_file = File(
        exists=True,
        argstr='--mask=%s',
        desc='filename for loading valid mask')
    nokspace = traits.Bool(
        False, argstr='--nokspace', desc='do not use k-space forward warping')

    # Special outputs: shift (voxel shift map, vsm)
    save_shift = traits.Bool(
        False, xor=['save_unmasked_shift'], desc='write pixel shift volume')
    shift_out_file = File(
        argstr='--saveshift=%s', desc='filename for saving pixel shift volume')
    save_unmasked_shift = traits.Bool(
        argstr='--unmaskshift',
        xor=['save_shift'],
        desc='saves the unmasked shiftmap when using --saveshift')

    # Special outputs: fieldmap (fmap)
    save_fmap = traits.Bool(
        False, xor=['save_unmasked_fmap'], desc='write field map volume')
    fmap_out_file = File(
        argstr='--savefmap=%s', desc='filename for saving fieldmap (rad/s)')
    save_unmasked_fmap = traits.Bool(
        False,
        argstr='--unmaskfmap',
        xor=['save_fmap'],
        desc='saves the unmasked fieldmap when using --savefmap')


class FUGUEOutputSpec(TraitedSpec):
    unwarped_file = File(desc='unwarped file')
    warped_file = File(desc='forward warped file')
    shift_out_file = File(desc='voxel shift map file')
    fmap_out_file = File(desc='fieldmap file')


class FUGUE(FSLCommand):
    """
    `FUGUE <http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FUGUE>`_ is, most generally,
    a set of tools for EPI distortion correction.

    Distortions may be corrected for
        1. improving registration with non-distorted images (e.g. structurals),
           or
        2. dealing with motion-dependent changes.

    FUGUE is designed to deal only with the first case -
    improving registration.


    Examples
    --------


    Unwarping an input image (shift map is known):

    >>> from nipype.interfaces.fsl.preprocess import FUGUE
    >>> fugue = FUGUE()
    >>> fugue.inputs.in_file = 'epi.nii'
    >>> fugue.inputs.mask_file = 'epi_mask.nii'
    >>> fugue.inputs.shift_in_file = 'vsm.nii'  # Previously computed with fugue as well
    >>> fugue.inputs.unwarp_direction = 'y'
    >>> fugue.inputs.output_type = "NIFTI_GZ"
    >>> fugue.cmdline # doctest: +ELLIPSIS
    'fugue --in=epi.nii --mask=epi_mask.nii --loadshift=vsm.nii --unwarpdir=y --unwarp=epi_unwarped.nii.gz'
    >>> fugue.run() #doctest: +SKIP


    Warping an input image (shift map is known):

    >>> from nipype.interfaces.fsl.preprocess import FUGUE
    >>> fugue = FUGUE()
    >>> fugue.inputs.in_file = 'epi.nii'
    >>> fugue.inputs.forward_warping = True
    >>> fugue.inputs.mask_file = 'epi_mask.nii'
    >>> fugue.inputs.shift_in_file = 'vsm.nii'  # Previously computed with fugue as well
    >>> fugue.inputs.unwarp_direction = 'y'
    >>> fugue.inputs.output_type = "NIFTI_GZ"
    >>> fugue.cmdline # doctest: +ELLIPSIS
    'fugue --in=epi.nii --mask=epi_mask.nii --loadshift=vsm.nii --unwarpdir=y --warp=epi_warped.nii.gz'
    >>> fugue.run() #doctest: +SKIP


    Computing the vsm (unwrapped phase map is known):

    >>> from nipype.interfaces.fsl.preprocess import FUGUE
    >>> fugue = FUGUE()
    >>> fugue.inputs.phasemap_in_file = 'epi_phasediff.nii'
    >>> fugue.inputs.mask_file = 'epi_mask.nii'
    >>> fugue.inputs.dwell_to_asym_ratio = (0.77e-3 * 3) / 2.46e-3
    >>> fugue.inputs.unwarp_direction = 'y'
    >>> fugue.inputs.save_shift = True
    >>> fugue.inputs.output_type = "NIFTI_GZ"
    >>> fugue.cmdline # doctest: +ELLIPSIS
    'fugue --dwelltoasym=0.9390243902 --mask=epi_mask.nii --phasemap=epi_phasediff.nii --saveshift=epi_phasediff_vsm.nii.gz --unwarpdir=y'
    >>> fugue.run() #doctest: +SKIP


    """

    _cmd = 'fugue'
    input_spec = FUGUEInputSpec
    output_spec = FUGUEOutputSpec

    def _parse_inputs(self, skip=None):
        if skip is None:
            skip = []

        input_phase = isdefined(self.inputs.phasemap_in_file)
        input_vsm = isdefined(self.inputs.shift_in_file)
        input_fmap = isdefined(self.inputs.fmap_in_file)

        if not input_phase and not input_vsm and not input_fmap:
            raise RuntimeError(
                ('Either phasemap_in_file, shift_in_file or fmap_in_file must '
                 'be set.'))

        if not isdefined(self.inputs.in_file):
            skip += ['unwarped_file', 'warped_file']
        else:
            if self.inputs.forward_warping:
                skip += ['unwarped_file']
                trait_spec = self.inputs.trait('warped_file')
                trait_spec.name_template = "%s_warped"
                trait_spec.name_source = 'in_file'
                trait_spec.output_name = 'warped_file'
            else:
                skip += ['warped_file']
                trait_spec = self.inputs.trait('unwarped_file')
                trait_spec.name_template = "%s_unwarped"
                trait_spec.name_source = 'in_file'
                trait_spec.output_name = 'unwarped_file'

        # Handle shift output
        if not isdefined(self.inputs.shift_out_file):
            vsm_save_masked = (isdefined(self.inputs.save_shift)
                               and self.inputs.save_shift)
            vsm_save_unmasked = (isdefined(self.inputs.save_unmasked_shift)
                                 and self.inputs.save_unmasked_shift)

            if (vsm_save_masked or vsm_save_unmasked):
                trait_spec = self.inputs.trait('shift_out_file')
                trait_spec.output_name = 'shift_out_file'

                if input_fmap:
                    trait_spec.name_source = 'fmap_in_file'
                elif input_phase:
                    trait_spec.name_source = 'phasemap_in_file'
                elif input_vsm:
                    trait_spec.name_source = 'shift_in_file'
                else:
                    raise RuntimeError(
                        ('Either phasemap_in_file, shift_in_file or '
                         'fmap_in_file must be set.'))

                if vsm_save_unmasked:
                    trait_spec.name_template = '%s_vsm_unmasked'
                else:
                    trait_spec.name_template = '%s_vsm'
            else:
                skip += ['save_shift', 'save_unmasked_shift', 'shift_out_file']

        # Handle fieldmap output
        if not isdefined(self.inputs.fmap_out_file):
            fmap_save_masked = (isdefined(self.inputs.save_fmap)
                                and self.inputs.save_fmap)
            fmap_save_unmasked = (isdefined(self.inputs.save_unmasked_fmap)
                                  and self.inputs.save_unmasked_fmap)

            if (fmap_save_masked or fmap_save_unmasked):
                trait_spec = self.inputs.trait('fmap_out_file')
                trait_spec.output_name = 'fmap_out_file'

                if input_vsm:
                    trait_spec.name_source = 'shift_in_file'
                elif input_phase:
                    trait_spec.name_source = 'phasemap_in_file'
                elif input_fmap:
                    trait_spec.name_source = 'fmap_in_file'
                else:
                    raise RuntimeError(
                        ('Either phasemap_in_file, shift_in_file or '
                         'fmap_in_file must be set.'))

                if fmap_save_unmasked:
                    trait_spec.name_template = '%s_fieldmap_unmasked'
                else:
                    trait_spec.name_template = '%s_fieldmap'
            else:
                skip += ['save_fmap', 'save_unmasked_fmap', 'fmap_out_file']

        return super(FUGUE, self)._parse_inputs(skip=skip)


class PRELUDEInputSpec(FSLCommandInputSpec):
    complex_phase_file = File(
        exists=True,
        argstr='--complex=%s',
        mandatory=True,
        xor=['magnitude_file', 'phase_file'],
        desc='complex phase input volume')
    magnitude_file = File(
        exists=True,
        argstr='--abs=%s',
        mandatory=True,
        xor=['complex_phase_file'],
        desc='file containing magnitude image')
    phase_file = File(
        exists=True,
        argstr='--phase=%s',
        mandatory=True,
        xor=['complex_phase_file'],
        desc='raw phase file')
    unwrapped_phase_file = File(
        genfile=True,
        argstr='--unwrap=%s',
        desc='file containing unwrapepd phase',
        hash_files=False)
    num_partitions = traits.Int(
        argstr='--numphasesplit=%d', desc='number of phase partitions to use')
    labelprocess2d = traits.Bool(
        argstr='--labelslices',
        desc='does label processing in 2D (slice at a time)')
    process2d = traits.Bool(
        argstr='--slices',
        xor=['labelprocess2d'],
        desc='does all processing in 2D (slice at a time)')
    process3d = traits.Bool(
        argstr='--force3D',
        xor=['labelprocess2d', 'process2d'],
        desc='forces all processing to be full 3D')
    threshold = traits.Float(
        argstr='--thresh=%.10f', desc='intensity threshold for masking')
    mask_file = File(
        exists=True, argstr='--mask=%s', desc='filename of mask input volume')
    start = traits.Int(
        argstr='--start=%d', desc='first image number to process (default 0)')
    end = traits.Int(
        argstr='--end=%d', desc='final image number to process (default Inf)')
    savemask_file = File(
        argstr='--savemask=%s',
        desc='saving the mask volume',
        hash_files=False)
    rawphase_file = File(
        argstr='--rawphase=%s',
        desc='saving the raw phase output',
        hash_files=False)
    label_file = File(
        argstr='--labels=%s',
        desc='saving the area labels output',
        hash_files=False)
    removeramps = traits.Bool(
        argstr='--removeramps', desc='remove phase ramps during unwrapping')


class PRELUDEOutputSpec(TraitedSpec):
    unwrapped_phase_file = File(exists=True, desc='unwrapped phase file')


class PRELUDE(FSLCommand):
    """Use FSL prelude to do phase unwrapping

    Examples
    --------

    Please insert examples for use of this command

    """
    input_spec = PRELUDEInputSpec
    output_spec = PRELUDEOutputSpec
    _cmd = 'prelude'

    def __init__(self, **kwargs):
        super(PRELUDE, self).__init__(**kwargs)
        warn('This has not been fully tested. Please report any failures.')

    def _list_outputs(self):
        outputs = self._outputs().get()
        out_file = self.inputs.unwrapped_phase_file
        if not isdefined(out_file):
            if isdefined(self.inputs.phase_file):
                out_file = self._gen_fname(
                    self.inputs.phase_file, suffix='_unwrapped')
            elif isdefined(self.inputs.complex_phase_file):
                out_file = self._gen_fname(
                    self.inputs.complex_phase_file, suffix='_phase_unwrapped')
        outputs['unwrapped_phase_file'] = os.path.abspath(out_file)
        return outputs

    def _gen_filename(self, name):
        if name == 'unwrapped_phase_file':
            return self._list_outputs()['unwrapped_phase_file']
        return None


class FIRSTInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        position=-2,
        copyfile=False,
        argstr='-i %s',
        desc='input data file')
    out_file = File(
        'segmented',
        usedefault=True,
        mandatory=True,
        position=-1,
        argstr='-o %s',
        desc='output data file',
        hash_files=False)
    verbose = traits.Bool(argstr='-v', position=1, desc="Use verbose logging.")
    brain_extracted = traits.Bool(
        argstr='-b',
        position=2,
        desc="Input structural image is already brain-extracted")
    no_cleanup = traits.Bool(
        argstr='-d',
        position=3,
        desc="Input structural image is already brain-extracted")
    method = traits.Enum(
        'auto',
        'fast',
        'none',
        xor=['method_as_numerical_threshold'],
        argstr='-m %s',
        position=4,
        usedefault=True,
        desc=("Method must be one of auto, fast, none, or it can be entered "
              "using the 'method_as_numerical_threshold' input"))
    method_as_numerical_threshold = traits.Float(
        argstr='-m %.4f',
        position=4,
        desc=("Specify a numerical threshold value or use the 'method' input "
              "to choose auto, fast, or none"))
    list_of_specific_structures = traits.List(
        traits.Str,
        argstr='-s %s',
        sep=',',
        position=5,
        minlen=1,
        desc='Runs only on the specified structures (e.g. L_Hipp, R_Hipp'
        'L_Accu, R_Accu, L_Amyg, R_Amyg'
        'L_Caud, R_Caud, L_Pall, R_Pall'
        'L_Puta, R_Puta, L_Thal, R_Thal, BrStem')
    affine_file = File(
        exists=True,
        position=6,
        argstr='-a %s',
        desc=('Affine matrix to use (e.g. img2std.mat) (does not '
              're-run registration)'))


class FIRSTOutputSpec(TraitedSpec):
    vtk_surfaces = OutputMultiPath(
        File(exists=True),
        desc='VTK format meshes for each subcortical region')
    bvars = OutputMultiPath(
        File(exists=True), desc='bvars for each subcortical region')
    original_segmentations = File(
        exists=True,
        desc=('3D image file containing the segmented regions '
              'as integer values. Uses CMA labelling'))
    segmentation_file = File(
        exists=True,
        desc=('4D image file containing a single volume per '
              'segmented region'))


class FIRST(FSLCommand):
    """
    Use FSL's run_first_all command to segment subcortical volumes

    http://www.fmrib.ox.ac.uk/fsl/first/index.html

    Examples
    --------

    >>> from nipype.interfaces import fsl
    >>> first = fsl.FIRST()
    >>> first.inputs.in_file = 'structural.nii'
    >>> first.inputs.out_file = 'segmented.nii'
    >>> res = first.run() #doctest: +SKIP

    """

    _cmd = 'run_first_all'
    input_spec = FIRSTInputSpec
    output_spec = FIRSTOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()

        if isdefined(self.inputs.list_of_specific_structures):
            structures = self.inputs.list_of_specific_structures
        else:
            structures = [
                'L_Hipp', 'R_Hipp', 'L_Accu', 'R_Accu', 'L_Amyg', 'R_Amyg',
                'L_Caud', 'R_Caud', 'L_Pall', 'R_Pall', 'L_Puta', 'R_Puta',
                'L_Thal', 'R_Thal', 'BrStem'
            ]
        outputs['original_segmentations'] = \
            self._gen_fname('original_segmentations')
        outputs['segmentation_file'] = self._gen_fname('segmentation_file')
        outputs['vtk_surfaces'] = self._gen_mesh_names('vtk_surfaces',
                                                       structures)
        outputs['bvars'] = self._gen_mesh_names('bvars', structures)
        return outputs

    def _gen_fname(self, basename):
        path, outname, ext = split_filename(self.inputs.out_file)

        method = 'none'
        if isdefined(self.inputs.method) and self.inputs.method != 'none':
            method = 'fast'
            if (self.inputs.list_of_specific_structures
                    and self.inputs.method == 'auto'):
                method = 'none'

        if isdefined(self.inputs.method_as_numerical_threshold):
            thres = '%.4f' % self.inputs.method_as_numerical_threshold
            method = thres.replace('.', '')

        if basename == 'original_segmentations':
            return op.abspath('%s_all_%s_origsegs.nii.gz' % (outname, method))
        if basename == 'segmentation_file':
            return op.abspath('%s_all_%s_firstseg.nii.gz' % (outname, method))

        return None

    def _gen_mesh_names(self, name, structures):
        path, prefix, ext = split_filename(self.inputs.out_file)
        if name == 'vtk_surfaces':
            vtks = list()
            for struct in structures:
                vtk = prefix + '-' + struct + '_first.vtk'
                vtks.append(op.abspath(vtk))
            return vtks
        if name == 'bvars':
            bvars = list()
            for struct in structures:
                bvar = prefix + '-' + struct + '_first.bvars'
                bvars.append(op.abspath(bvar))
            return bvars
        return None



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..registration import scalartransform


def test_scalartransform_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        deformation=dict(argstr='--deformation %s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        h_field=dict(argstr='--h_field ', ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        input_image=dict(argstr='--input_image %s', ),
        interpolation=dict(argstr='--interpolation %s', ),
        invert=dict(argstr='--invert ', ),
        output_image=dict(
            argstr='--output_image %s',
            hash_files=False,
        ),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
        transformation=dict(
            argstr='--transformation %s',
            hash_files=False,
        ),
    )
    inputs = scalartransform.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_scalartransform_outputs():
    output_map = dict(
        output_image=dict(),
        transformation=dict(),
    )
    outputs = scalartransform.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..connectivity import LabelConfig


def test_LabelConfig_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        in_config=dict(
            argstr='%s',
            position=-2,
        ),
        in_file=dict(
            argstr='%s',
            mandatory=True,
            position=-3,
        ),
        lut_aal=dict(argstr='-lut_aal %s', ),
        lut_basic=dict(argstr='-lut_basic %s', ),
        lut_fs=dict(argstr='-lut_freesurfer %s', ),
        lut_itksnap=dict(argstr='-lut_itksnap %s', ),
        nthreads=dict(
            argstr='-nthreads %d',
            nohash=True,
        ),
        out_file=dict(
            argstr='%s',
            mandatory=True,
            position=-1,
            usedefault=True,
        ),
        spine=dict(argstr='-spine %s', ),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
    )
    inputs = LabelConfig.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_LabelConfig_outputs():
    output_map = dict(out_file=dict(), )
    outputs = LabelConfig.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
import os.path as op  # system functions

from .connectivity_mapping import create_connectivity_pipeline
from ....interfaces import io as nio  # Data i/o
from ....interfaces import utility as util  # utility
from ....pipeline import engine as pe  # pypeline engine


def create_group_connectivity_pipeline(group_list,
                                       group_id,
                                       data_dir,
                                       subjects_dir,
                                       output_dir,
                                       template_args_dict=0):
    """Creates a pipeline that performs basic Camino structural connectivity processing
    on groups of subjects. Given a diffusion-weighted image, and text files containing
    the associated b-values and b-vectors, the workflow will return each subjects' connectomes
    in a Connectome File Format (CFF) file, for use in Connectome Viewer (http://www.cmtk.org).

    Example
    -------

    >>> import nipype.interfaces.freesurfer as fs
    >>> import nipype.workflows.dmri.camino.group_connectivity as groupwork
    >>> subjects_dir = '.'
    >>> data_dir = '.'
    >>> output_dir = '.'
    >>> fs.FSCommand.set_default_subjects_dir(subjects_dir)
    >>> group_list = {}
    >>> group_list['group1'] = ['subj1', 'subj2']
    >>> group_list['group2'] = ['subj3', 'subj4']
    >>> template_args = dict(dwi=[['subject_id', 'dwi']], bvecs=[['subject_id', 'bvecs']], bvals=[['subject_id', 'bvals']])
    >>> group_id = 'group1'
    >>> l1pipeline = groupwork.create_group_connectivity_pipeline(group_list, group_id, data_dir, subjects_dir, output_dir, template_args)
    >>> l1pipeline.run()                 # doctest: +SKIP

    Inputs::

        group_list: Dictionary of subject lists, keyed by group name
        group_id: String containing the group name
        data_dir: Path to the data directory
        subjects_dir: Path to the Freesurfer 'subjects' directory
        output_dir: Path for the output files
        template_args_dict: Dictionary of template arguments for the connectivity pipeline datasource
                                e.g.    info = dict(dwi=[['subject_id', 'dwi']],
                                                bvecs=[['subject_id','bvecs']],
                                                bvals=[['subject_id','bvals']])
    """
    group_infosource = pe.Node(
        interface=util.IdentityInterface(fields=['group_id']),
        name="group_infosource")
    group_infosource.inputs.group_id = group_id
    subject_list = group_list[group_id]
    subj_infosource = pe.Node(
        interface=util.IdentityInterface(fields=['subject_id']),
        name="subj_infosource")
    subj_infosource.iterables = ('subject_id', subject_list)

    if template_args_dict == 0:
        info = dict(
            dwi=[['subject_id', 'dwi']],
            bvecs=[['subject_id', 'bvecs']],
            bvals=[['subject_id', 'bvals']])
    else:
        info = template_args_dict

    datasource = pe.Node(
        interface=nio.DataGrabber(
            infields=['subject_id'], outfields=list(info.keys())),
        name='datasource')

    datasource.inputs.template = "%s/%s"
    datasource.inputs.base_directory = data_dir
    datasource.inputs.field_template = dict(dwi='%s/%s.nii')
    datasource.inputs.template_args = info
    datasource.inputs.sort_filelist = True
    """
    Create a connectivity mapping workflow
    """
    conmapper = create_connectivity_pipeline("nipype_conmap")
    conmapper.inputs.inputnode.subjects_dir = subjects_dir
    conmapper.base_dir = op.abspath('conmapper')

    datasink = pe.Node(interface=nio.DataSink(), name="datasink")
    datasink.inputs.base_directory = output_dir
    datasink.inputs.container = group_id

    l1pipeline = pe.Workflow(name="l1pipeline_" + group_id)
    l1pipeline.base_dir = output_dir
    l1pipeline.base_output_dir = group_id
    l1pipeline.connect([(subj_infosource, datasource, [('subject_id',
                                                        'subject_id')])])
    l1pipeline.connect([(subj_infosource, conmapper,
                         [('subject_id', 'inputnode.subject_id')])])
    l1pipeline.connect([(datasource, conmapper, [
        ("dwi", "inputnode.dwi"),
        ("bvals", "inputnode.bvals"),
        ("bvecs", "inputnode.bvecs"),
    ])])
    l1pipeline.connect([(conmapper, datasink, [
        ("outputnode.connectome", "@l1output.cff"),
        ("outputnode.fa", "@l1output.fa"),
        ("outputnode.tracts", "@l1output.tracts"),
        ("outputnode.trace", "@l1output.trace"),
        ("outputnode.cmatrix", "@l1output.cmatrix"),
        ("outputnode.rois", "@l1output.rois"),
        ("outputnode.struct", "@l1output.struct"),
        ("outputnode.networks", "@l1output.networks"),
        ("outputnode.mean_fiber_length", "@l1output.mean_fiber_length"),
        ("outputnode.fiber_length_std", "@l1output.fiber_length_std"),
    ])])
    l1pipeline.connect([(group_infosource, datasink, [('group_id',
                                                       '@group_id')])])
    return l1pipeline



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from builtins import range

from ....interfaces import fsl as fsl  # fsl
from ....interfaces import utility as util  # utility
from ....pipeline import engine as pe  # pypeline engine

from .... import LooseVersion


def create_modelfit_workflow(name='modelfit', f_contrasts=False):
    """Create an FSL individual modelfitting workflow

    Example
    -------

    >>> modelfit = create_modelfit_workflow()
    >>> modelfit.base_dir = '.'
    >>> info = dict()
    >>> modelfit.inputs.inputspec.session_info = info
    >>> modelfit.inputs.inputspec.interscan_interval = 3.
    >>> modelfit.inputs.inputspec.film_threshold = 1000
    >>> modelfit.run() #doctest: +SKIP

    Inputs::

         inputspec.session_info : info generated by modelgen.SpecifyModel
         inputspec.interscan_interval : interscan interval
         inputspec.contrasts : list of contrasts
         inputspec.film_threshold : image threshold for FILM estimation
         inputspec.model_serial_correlations
         inputspec.bases

    Outputs::

         outputspec.copes
         outputspec.varcopes
         outputspec.dof_file
         outputspec.pfiles
         outputspec.zfiles
         outputspec.parameter_estimates
    """

    version = 0
    if fsl.Info.version() and \
            LooseVersion(fsl.Info.version()) > LooseVersion('5.0.6'):
        version = 507

    modelfit = pe.Workflow(name=name)
    """
    Create the nodes
    """

    inputspec = pe.Node(
        util.IdentityInterface(fields=[
            'session_info', 'interscan_interval', 'contrasts',
            'film_threshold', 'functional_data', 'bases',
            'model_serial_correlations'
        ]),
        name='inputspec')
    level1design = pe.Node(interface=fsl.Level1Design(), name="level1design")
    modelgen = pe.MapNode(
        interface=fsl.FEATModel(),
        name='modelgen',
        iterfield=['fsf_file', 'ev_files'])
    if version < 507:
        modelestimate = pe.MapNode(
            interface=fsl.FILMGLS(smooth_autocorr=True, mask_size=5),
            name='modelestimate',
            iterfield=['design_file', 'in_file'])
    else:
        if f_contrasts:
            iterfield = ['design_file', 'in_file', 'tcon_file', 'fcon_file']
        else:
            iterfield = ['design_file', 'in_file', 'tcon_file']
        modelestimate = pe.MapNode(
            interface=fsl.FILMGLS(smooth_autocorr=True, mask_size=5),
            name='modelestimate',
            iterfield=iterfield)

    if version < 507:
        if f_contrasts:
            iterfield = [
                'tcon_file', 'fcon_file', 'param_estimates', 'sigmasquareds',
                'corrections', 'dof_file'
            ]
        else:
            iterfield = [
                'tcon_file', 'param_estimates', 'sigmasquareds', 'corrections',
                'dof_file'
            ]
        conestimate = pe.MapNode(
            interface=fsl.ContrastMgr(),
            name='conestimate',
            iterfield=[
                'tcon_file', 'fcon_file', 'param_estimates', 'sigmasquareds',
                'corrections', 'dof_file'
            ])

    if f_contrasts:
        iterfield = ['in1', 'in2']
    else:
        iterfield = ['in1']
    merge_contrasts = pe.MapNode(
        interface=util.Merge(2), name='merge_contrasts', iterfield=iterfield)
    ztopval = pe.MapNode(
        interface=fsl.ImageMaths(op_string='-ztop', suffix='_pval'),
        nested=True,
        name='ztop',
        iterfield=['in_file'])
    outputspec = pe.Node(
        util.IdentityInterface(fields=[
            'copes', 'varcopes', 'dof_file', 'pfiles', 'zfiles',
            'parameter_estimates'
        ]),
        name='outputspec')
    """
    Setup the connections
    """

    modelfit.connect([
        (inputspec, level1design,
         [('interscan_interval', 'interscan_interval'),
          ('session_info', 'session_info'), ('contrasts', 'contrasts'),
          ('bases', 'bases'), ('model_serial_correlations',
                               'model_serial_correlations')]),
        (inputspec, modelestimate, [('film_threshold', 'threshold'),
                                    ('functional_data', 'in_file')]),
        (level1design, modelgen, [('fsf_files', 'fsf_file'), ('ev_files',
                                                              'ev_files')]),
        (modelgen, modelestimate, [('design_file', 'design_file')]),
        (merge_contrasts, ztopval, [('out', 'in_file')]),
        (ztopval, outputspec, [('out_file', 'pfiles')]),
        (merge_contrasts, outputspec, [('out', 'zfiles')]),
        (modelestimate, outputspec, [('param_estimates',
                                      'parameter_estimates'), ('dof_file',
                                                               'dof_file')]),
    ])
    if version < 507:
        modelfit.connect([
            (modelgen, conestimate, [('con_file', 'tcon_file'),
                                     ('fcon_file', 'fcon_file')]),
            (modelestimate, conestimate,
             [('param_estimates', 'param_estimates'), ('sigmasquareds',
                                                       'sigmasquareds'),
              ('corrections', 'corrections'), ('dof_file', 'dof_file')]),
            (conestimate, merge_contrasts, [('zstats', 'in1'), ('zfstats',
                                                                'in2')]),
            (conestimate, outputspec, [('copes', 'copes'), ('varcopes',
                                                            'varcopes')]),
        ])
    else:
        modelfit.connect([
            (modelgen, modelestimate, [('con_file', 'tcon_file'),
                                       ('fcon_file', 'fcon_file')]),
            (modelestimate, merge_contrasts, [('zstats', 'in1'), ('zfstats',
                                                                  'in2')]),
            (modelestimate, outputspec, [('copes', 'copes'), ('varcopes',
                                                              'varcopes')]),
        ])
    return modelfit


def create_overlay_workflow(name='overlay'):
    """Setup overlay workflow
    """

    overlay = pe.Workflow(name='overlay')
    overlaystats = pe.MapNode(
        interface=fsl.Overlay(), name="overlaystats", iterfield=['stat_image'])
    overlaystats.inputs.show_negative_stats = True
    overlaystats.inputs.auto_thresh_bg = True

    slicestats = pe.MapNode(
        interface=fsl.Slicer(), name="slicestats", iterfield=['in_file'])
    slicestats.inputs.all_axial = True
    slicestats.inputs.image_width = 512

    overlay.connect(overlaystats, 'out_file', slicestats, 'in_file')
    return overlay


def create_fixed_effects_flow(name='fixedfx'):
    """Create a fixed-effects workflow

    This workflow is used to combine registered copes and varcopes across runs
    for an individual subject

    Example
    -------

    >>> fixedfx = create_fixed_effects_flow()
    >>> fixedfx.base_dir = '.'
    >>> fixedfx.inputs.inputspec.copes = [['cope1run1.nii.gz', 'cope1run2.nii.gz'], ['cope2run1.nii.gz', 'cope2run2.nii.gz']] # per contrast
    >>> fixedfx.inputs.inputspec.varcopes = [['varcope1run1.nii.gz', 'varcope1run2.nii.gz'], ['varcope2run1.nii.gz', 'varcope2run2.nii.gz']] # per contrast
    >>> fixedfx.inputs.inputspec.dof_files = ['dofrun1', 'dofrun2'] # per run
    >>> fixedfx.run() #doctest: +SKIP

    Inputs::

         inputspec.copes : list of list of cope files (one list per contrast)
         inputspec.varcopes : list of list of varcope files (one list per
                              contrast)
         inputspec.dof_files : degrees of freedom files for each run

    Outputs::

         outputspec.res4d : 4d residual time series
         outputspec.copes : contrast parameter estimates
         outputspec.varcopes : variance of contrast parameter estimates
         outputspec.zstats : z statistics of contrasts
         outputspec.tstats : t statistics of contrasts
    """

    fixed_fx = pe.Workflow(name=name)

    inputspec = pe.Node(
        util.IdentityInterface(fields=['copes', 'varcopes', 'dof_files']),
        name='inputspec')
    """
    Use :class:`nipype.interfaces.fsl.Merge` to merge the copes and
    varcopes for each condition
    """

    copemerge = pe.MapNode(
        interface=fsl.Merge(dimension='t'),
        iterfield=['in_files'],
        name="copemerge")

    varcopemerge = pe.MapNode(
        interface=fsl.Merge(dimension='t'),
        iterfield=['in_files'],
        name="varcopemerge")
    """
    Use :class:`nipype.interfaces.fsl.L2Model` to generate subject and condition
    specific level 2 model design files
    """

    level2model = pe.Node(interface=fsl.L2Model(), name='l2model')
    """
    Use :class:`nipype.interfaces.fsl.FLAMEO` to estimate a second level model
    """

    flameo = pe.MapNode(
        interface=fsl.FLAMEO(run_mode='fe'),
        name="flameo",
        iterfield=['cope_file', 'var_cope_file'])

    def get_dofvolumes(dof_files, cope_files):
        import os
        import nibabel as nb
        import numpy as np
        img = nb.load(cope_files[0])
        if len(img.shape) > 3:
            out_data = np.zeros(img.shape)
        else:
            out_data = np.zeros(list(img.shape) + [1])
        for i in range(out_data.shape[-1]):
            dof = np.loadtxt(dof_files[i])
            out_data[:, :, :, i] = dof
        filename = os.path.join(os.getcwd(), 'dof_file.nii.gz')
        newimg = nb.Nifti1Image(out_data, None, img.header)
        newimg.to_filename(filename)
        return filename

    gendof = pe.Node(
        util.Function(
            input_names=['dof_files', 'cope_files'],
            output_names=['dof_volume'],
            function=get_dofvolumes),
        name='gendofvolume')

    outputspec = pe.Node(
        util.IdentityInterface(
            fields=['res4d', 'copes', 'varcopes', 'zstats', 'tstats']),
        name='outputspec')

    fixed_fx.connect(
        [(inputspec, copemerge,
          [('copes', 'in_files')]), (inputspec, varcopemerge, [('varcopes',
                                                                'in_files')]),
         (inputspec, gendof, [('dof_files', 'dof_files')]), (copemerge, gendof,
                                                             [('merged_file',
                                                               'cope_files')]),
         (copemerge, flameo,
          [('merged_file', 'cope_file')]), (varcopemerge, flameo, [
              ('merged_file', 'var_cope_file')
          ]), (level2model, flameo,
               [('design_mat', 'design_file'), ('design_con', 't_con_file'),
                ('design_grp', 'cov_split_file')]), (gendof, flameo,
                                                     [('dof_volume',
                                                       'dof_var_cope_file')]),
         (flameo, outputspec,
          [('res4d', 'res4d'), ('copes', 'copes'), ('var_copes', 'varcopes'),
           ('zstats', 'zstats'), ('tstats', 'tstats')])])
    return fixed_fx
